import sys

import numpy as np
import pytest


def test_camera(make_napari_viewer):
    """Test vispy camera creation in 2D."""
    viewer = make_napari_viewer()
    vispy_camera = viewer.window._qt_viewer.canvas.camera

    np.random.seed(0)
    data = np.random.random((11, 11, 11))
    viewer.add_image(data)

    # Test default values camera values are used and vispy camera has been
    # updated
    assert viewer.dims.ndisplay == 2

    np.testing.assert_almost_equal(viewer.camera.angles, (0, 0, 90))
    np.testing.assert_almost_equal(viewer.camera.center, (0, 5.0, 5.0))
    np.testing.assert_almost_equal(viewer.camera.angles, vispy_camera.angles)
    np.testing.assert_almost_equal(viewer.camera.center, vispy_camera.center)
    np.testing.assert_almost_equal(viewer.camera.zoom, vispy_camera.zoom)


def test_vispy_camera_update_from_model(make_napari_viewer):
    """Test vispy camera update from model in 2D."""
    viewer = make_napari_viewer()
    vispy_camera = viewer.window._qt_viewer.canvas.camera

    np.random.seed(0)
    data = np.random.random((11, 11, 11))
    viewer.add_image(data)

    # Test default values camera values are used and vispy camera has been
    # updated
    assert viewer.dims.ndisplay == 2

    # Update camera center and zoom
    viewer.camera.center = (11, 12)
    viewer.camera.zoom = 4

    np.testing.assert_almost_equal(viewer.camera.angles, (0, 0, 90))
    np.testing.assert_almost_equal(viewer.camera.center, (0, 11, 12))
    np.testing.assert_almost_equal(viewer.camera.zoom, 4)
    np.testing.assert_almost_equal(viewer.camera.angles, vispy_camera.angles)
    np.testing.assert_almost_equal(viewer.camera.center, vispy_camera.center)
    np.testing.assert_almost_equal(viewer.camera.zoom, vispy_camera.zoom)


def test_camera_model_update_from_vispy(make_napari_viewer):
    """Test camera model updates from vispy in 2D."""
    viewer = make_napari_viewer()
    vispy_camera = viewer.window._qt_viewer.canvas.camera

    np.random.seed(0)
    data = np.random.random((11, 11, 11))
    viewer.add_image(data)

    # Test default values camera values are used and vispy camera has been
    # updated
    assert viewer.dims.ndisplay == 2

    # Update vispy camera center and zoom
    vispy_camera.center = (11, 12)
    vispy_camera.zoom = 4
    vispy_camera.on_draw(None)

    np.testing.assert_almost_equal(viewer.camera.angles, (0, 0, 90))
    np.testing.assert_almost_equal(viewer.camera.center, (0, 11, 12))
    np.testing.assert_almost_equal(viewer.camera.zoom, 4)
    np.testing.assert_almost_equal(viewer.camera.angles, vispy_camera.angles)
    np.testing.assert_almost_equal(viewer.camera.center, vispy_camera.center)
    np.testing.assert_almost_equal(viewer.camera.zoom, vispy_camera.zoom)


def test_3D_camera(make_napari_viewer):
    """Test vispy camera creation in 3D."""
    viewer = make_napari_viewer()
    vispy_camera = viewer.window._qt_viewer.canvas.camera

    np.random.seed(0)
    data = np.random.random((11, 11, 11))
    viewer.add_image(data)

    viewer.dims.ndisplay = 3

    # Test camera values have updated
    np.testing.assert_almost_equal(viewer.camera.angles, (0, 0, 90))
    np.testing.assert_almost_equal(viewer.camera.center, (5.0, 5.0, 5.0))
    np.testing.assert_almost_equal(viewer.camera.angles, vispy_camera.angles)
    np.testing.assert_almost_equal(viewer.camera.center, vispy_camera.center)
    np.testing.assert_almost_equal(viewer.camera.zoom, vispy_camera.zoom)


def test_vispy_camera_update_from_model_3D(make_napari_viewer):
    """Test vispy camera update from model in 3D."""
    viewer = make_napari_viewer()
    vispy_camera = viewer.window._qt_viewer.canvas.camera

    np.random.seed(0)
    data = np.random.random((11, 11, 11))
    viewer.add_image(data)

    viewer.dims.ndisplay = 3

    # Update camera angles, center, and zoom
    viewer.camera.angles = (24, 12, -19)
    viewer.camera.center = (11, 12, 15)
    viewer.camera.zoom = 4

    np.testing.assert_almost_equal(viewer.camera.angles, (24, 12, -19))
    np.testing.assert_almost_equal(viewer.camera.center, (11, 12, 15))
    np.testing.assert_almost_equal(viewer.camera.zoom, 4)
    np.testing.assert_almost_equal(viewer.camera.angles, vispy_camera.angles)
    np.testing.assert_almost_equal(viewer.camera.center, vispy_camera.center)
    np.testing.assert_almost_equal(viewer.camera.zoom, vispy_camera.zoom)


def test_camera_model_update_from_vispy_3D(make_napari_viewer):
    """Test camera model updates from vispy in 3D."""
    viewer = make_napari_viewer()
    vispy_camera = viewer.window._qt_viewer.canvas.camera

    np.random.seed(0)
    data = np.random.random((11, 11, 11))
    viewer.add_image(data)

    viewer.dims.ndisplay = 3

    # Update vispy camera angles, center, and zoom
    viewer.camera.angles = (24, 12, -19)
    vispy_camera.center = (11, 12, 15)
    vispy_camera.zoom = 4
    vispy_camera.on_draw(None)

    np.testing.assert_almost_equal(viewer.camera.angles, (24, 12, -19))
    np.testing.assert_almost_equal(viewer.camera.center, (11, 12, 15))
    np.testing.assert_almost_equal(viewer.camera.zoom, 4)
    np.testing.assert_almost_equal(viewer.camera.angles, vispy_camera.angles)
    np.testing.assert_almost_equal(viewer.camera.center, vispy_camera.center)
    np.testing.assert_almost_equal(viewer.camera.zoom, vispy_camera.zoom)


def test_switching_ndisplay_maintains_3D_angles(make_napari_viewer):
    """Test that switching dims.ndisplay maintains 3D angles."""
    viewer = make_napari_viewer()
    vispy_camera = viewer.window._qt_viewer.canvas.camera

    np.random.seed(0)
    data = np.random.random((11, 11, 11))
    viewer.add_image(data)

    angles_3D = (24, 12, -19)
    angles_2D = (0, 0, 90)

    viewer.dims.ndisplay = 3
    viewer.camera.angles = angles_3D
    np.testing.assert_almost_equal(viewer.camera.angles, vispy_camera.angles)

    # switching to 2D should maintain the model camera angles from 3D
    # but the vispy camera angles will be the default 2D angles
    viewer.dims.ndisplay = 2
    np.testing.assert_almost_equal(viewer.camera.angles, angles_3D)
    np.testing.assert_almost_equal(vispy_camera.angles, angles_2D)

    # switching back to 3D should use the model camera angles for the
    # vispy camera
    viewer.dims.ndisplay = 3
    np.testing.assert_almost_equal(viewer.camera.angles, vispy_camera.angles)


@pytest.mark.skipif(
    sys.platform == 'win32', reason='This new test is flaky on windows'
)
def test_camera_orientation_2d(make_napari_viewer, qtbot):
    """Test that flipping orientation of the camera flips displayed image."""
    viewer = make_napari_viewer(show=True)
    data = np.arange(16).reshape((4, 4))
    _ = viewer.add_image(data, interpolation2d='linear')

    # in the default axis orientation of (down, right), the values in a
    # screenshot should continually increase as you go down in the image.
    # We take only the first channel in the RGBA array for simplicity, since
    # this is a grayscale image.
    qtbot.wait(50)
    sshot0 = viewer.window.export_figure(scale=10, flash=False)[..., 0]
    # check that the values are monotonically increasing down:
    avg_row_intensity_grad0 = np.diff(np.mean(sshot0, axis=1))
    assert np.all(avg_row_intensity_grad0 >= 0)

    # same but to the right
    avg_col_intensity_grad0 = np.diff(np.mean(sshot0, axis=0))
    assert np.all(avg_col_intensity_grad0 >= 0)

    # now we reverse the orientation of the vertical axis, and check that the
    # row gradient has changed direction but not the col gradient
    viewer.camera.orientation2d = ('up', 'right')
    qtbot.wait(50)
    sshot1 = viewer.window.export_figure(scale=10, flash=False)[..., 0]
    avg_row_intensity_grad1 = np.diff(np.mean(sshot1, axis=1))
    assert np.all(avg_row_intensity_grad1 <= 0)  # note inverted sign
    avg_col_intensity_grad1 = np.diff(np.mean(sshot1, axis=0))
    assert np.all(avg_col_intensity_grad1 >= 0)

    # finally, reverse orientation of horizontal axis, check that col gradient
    # has now also changed direction
    viewer.camera.orientation2d = ('up', 'left')
    qtbot.wait(50)
    sshot2 = viewer.window.export_figure(scale=10, flash=False)[..., 0]
    avg_row_intensity_grad2 = np.diff(np.mean(sshot2, axis=1))
    assert np.all(avg_row_intensity_grad2 <= 0)  # note inverted sign
    avg_col_intensity_grad2 = np.diff(np.mean(sshot2, axis=0))
    assert np.all(avg_col_intensity_grad2 <= 0)


@pytest.mark.skipif(
    sys.platform == 'win32', reason='This new test is flaky on windows'
)
def test_camera_orientation_3d(make_napari_viewer, qtbot):
    """Test that flipping camera orientation in 3D flips volume as expected."""
    viewer = make_napari_viewer(show=True)
    viewer.dims.ndisplay = 3
    gradient_z = np.arange(16).reshape((16, 1, 1))
    image = np.ones((16, 16))
    image_3d = gradient_z * image
    _ = viewer.add_image(image_3d)

    # We have a 3D image with brightness *increasing with z* (the 0th axis).
    # Therefore, when using a perspective projection, if we point z *away* from
    # us, the bright part of the image will be small, whereas if we point z
    # *towards* us, it will be large. We test the dimension flip by comparing
    # the overall brightness of the image when pointing z in different
    # directions

    viewer.camera.perspective = 60
    viewer.camera.orientation = ('away', 'down', 'right')
    qtbot.wait(50)
    sshot_away = viewer.screenshot(canvas_only=True, flash=False)[..., 0]
    viewer.camera.orientation = ('towards', 'down', 'right')
    qtbot.wait(50)
    sshot_towards = viewer.screenshot(canvas_only=True, flash=False)[..., 0]

    assert np.mean(sshot_towards) > np.mean(sshot_away)



import pytest

from napari.layers.points import Points, _points_key_bindings as key_bindings


@pytest.mark.key_bindings
def test_modes(layer):
    data = [[1, 3], [8, 4], [10, 10], [15, 4]]
    layer = Points(data, size=1)

    key_bindings.activate_points_add_mode(layer)
    assert layer.mode == 'add'
    key_bindings.activate_points_select_mode(layer)
    assert layer.mode == 'select'
    key_bindings.activate_points_pan_zoom_mode(layer)
    assert layer.mode == 'pan_zoom'


@pytest.mark.key_bindings
def test_copy_paste(layer):
    data = [[1, 3], [8, 4], [10, 10], [15, 4]]
    layer = Points(data, size=1)
    layer.mode = 'select'

    assert len(layer.data) == 4
    assert layer._clipboard == {}
    layer.selected_data = {0, 1}

    key_bindings.copy(layer)
    assert len(layer.data) == 4
    assert len(layer._clipboard) > 0

    key_bindings.paste(layer)
    assert len(layer.data) == 6
    assert len(layer._clipboard) > 0


@pytest.mark.key_bindings
def test_select_all_in_slice(layer):
    data = [[1, 3], [8, 4], [10, 10], [15, 4]]
    layer = Points(data, size=1)
    layer.mode = 'select'
    layer._set_view_slice()

    assert len(layer.data) == 4
    assert len(layer.selected_data) == 0

    key_bindings.select_all_in_slice(layer)
    assert len(layer.selected_data) == 4

    key_bindings.select_all_in_slice(layer)
    assert len(layer.selected_data) == 0


@pytest.mark.key_bindings
def test_select_all_in_slice_3d_data(layer):
    data = [[0, 1, 3], [0, 8, 4], [0, 10, 10], [1, 15, 4]]
    layer = Points(data, size=1)
    layer.mode = 'select'
    layer._set_view_slice()

    assert len(layer.data) == 4
    assert len(layer.selected_data) == 0

    key_bindings.select_all_in_slice(layer)
    assert len(layer.selected_data) == 3

    key_bindings.select_all_in_slice(layer)
    assert len(layer.selected_data) == 0


@pytest.mark.key_bindings
def test_select_all_data(layer):
    data = [[1, 3], [8, 4], [10, 10], [15, 4]]
    layer = Points(data, size=1)
    layer.mode = 'select'
    layer._set_view_slice()

    assert len(layer.data) == 4
    assert len(layer.selected_data) == 0

    key_bindings.select_all_data(layer)
    assert len(layer.selected_data) == 4

    key_bindings.select_all_data(layer)
    assert len(layer.selected_data) == 0


@pytest.mark.key_bindings
def test_select_all_data_3d_data(layer):
    data = [[0, 1, 3], [0, 8, 4], [0, 10, 10], [1, 15, 4]]
    layer = Points(data, size=1)
    layer.mode = 'select'
    layer._set_view_slice()

    assert len(layer.data) == 4
    assert len(layer.selected_data) == 0

    key_bindings.select_all_data(layer)
    assert len(layer.selected_data) == 4

    key_bindings.select_all_data(layer)
    assert len(layer.selected_data) == 0


def test_select_all_mixed(layer):
    data = [[0, 1, 3], [0, 8, 4], [0, 10, 10], [1, 15, 4]]
    layer = Points(data, size=1)
    layer.mode = 'select'
    layer._set_view_slice()

    assert len(layer.data) == 4
    assert len(layer.selected_data) == 0

    key_bindings.select_all_data(layer)
    assert len(layer.selected_data) == 4

    key_bindings.select_all_in_slice(layer)
    assert len(layer.selected_data) == 1

    key_bindings.select_all_in_slice(layer)
    assert len(layer.selected_data) == 4

    key_bindings.select_all_in_slice(layer)
    assert len(layer.selected_data) == 1

    key_bindings.select_all_data(layer)
    assert len(layer.selected_data) == 4

    key_bindings.select_all_data(layer)
    assert len(layer.selected_data) == 0


def test_delete_selected_points(layer):
    data = [[1, 3], [8, 4], [10, 10], [15, 4]]
    layer = Points(data, size=1)
    layer.mode = 'select'

    assert len(layer.data) == 4
    layer.selected_data = {0, 1}

    key_bindings.delete_selected_points(layer)
    assert len(layer.data) == 2



from unittest.mock import patch

import numpy as np
import pytest
from qtpy.QtCore import QObject, Signal
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QApplication, QColorDialog, QMainWindow

from napari._qt.utils import (
    QBYTE_FLAG,
    add_flash_animation,
    get_color,
    is_qbyte,
    qbytearray_to_str,
    qt_might_be_rich_text,
    qt_signals_blocked,
    str_to_qbytearray,
)
from napari.utils._proxies import PublicOnlyProxy


class Emitter(QObject):
    test_signal = Signal()

    def go(self):
        self.test_signal.emit()


def test_signal_blocker(qtbot):
    """make sure context manager signal blocker works"""
    import pytestqt.exceptions

    obj = Emitter()

    # make sure signal works
    with qtbot.waitSignal(obj.test_signal):
        obj.go()

    # make sure blocker works
    with qt_signals_blocked(obj):
        with pytest.raises(pytestqt.exceptions.TimeoutError):
            with qtbot.waitSignal(obj.test_signal, timeout=500):
                obj.go()


def test_is_qbyte_valid():
    is_qbyte(QBYTE_FLAG)
    is_qbyte(
        '!QBYTE_AAAA/wAAAAD9AAAAAgAAAAAAAAECAAACePwCAAAAAvsAAAAcAGwAYQB5AGUAcgAgAGMAbwBuAHQAcgBvAGwAcwEAAAAAAAABFwAAARcAAAEX+wAAABQAbABhAHkAZQByACAAbABpAHMAdAEAAAEXAAABYQAAALcA////AAAAAwAAAAAAAAAA/AEAAAAB+wAAAA4AYwBvAG4AcwBvAGwAZQAAAAAA/////wAAADIA////AAADPAAAAngAAAAEAAAABAAAAAgAAAAI/AAAAAA='
    )


def test_str_to_qbytearray_valid():
    str_to_qbytearray(
        '!QBYTE_AAAA/wAAAAD9AAAAAgAAAAAAAAECAAACePwCAAAAAvsAAAAcAGwAYQB5AGUAcgAgAGMAbwBuAHQAcgBvAGwAcwEAAAAAAAABFwAAARcAAAEX+wAAABQAbABhAHkAZQByACAAbABpAHMAdAEAAAEXAAABYQAAALcA////AAAAAwAAAAAAAAAA/AEAAAAB+wAAAA4AYwBvAG4AcwBvAGwAZQAAAAAA/////wAAADIA////AAADPAAAAngAAAAEAAAABAAAAAgAAAAI/AAAAAA='
    )


def test_str_to_qbytearray_invalid():
    with pytest.raises(ValueError, match='Invalid QByte string.'):
        str_to_qbytearray('')

    with pytest.raises(ValueError, match='Invalid QByte string.'):
        str_to_qbytearray('FOOBAR')

    with pytest.raises(ValueError, match='Invalid QByte string.'):
        str_to_qbytearray(
            '_AAAA/wAAAAD9AAAAAgAAAAAAAAECAAACePwCAAAAAvsAAAAcAGwAYQB5AGUAcgAgAGMAbwBuAHQAcgBvAGwAcwEAAAAAAAABFwAAARcAAAEX+wAAABQAbABhAHkAZQByACAAbABpAHMAdAEAAAEXAAABYQAAALcA////AAAAAwAAAAAAAAAA/AEAAAAB+wAAAA4AYwBvAG4AcwBvAGwAZQAAAAAA/////wAAADIA////AAADPAAAAngAAAAEAAAABAAAAAgAAAAI/AAAAAA='
        )


def test_qbytearray_to_str(qtbot):
    widget = QMainWindow()
    qtbot.addWidget(widget)

    qbyte = widget.saveState()
    qbyte_string = qbytearray_to_str(qbyte)
    assert is_qbyte(qbyte_string)


def test_qbytearray_to_str_and_back(qtbot):
    widget = QMainWindow()
    qtbot.addWidget(widget)

    qbyte = widget.saveState()
    assert str_to_qbytearray(qbytearray_to_str(qbyte)) == qbyte


def test_add_flash_animation(qtbot):
    widget = QMainWindow()
    qtbot.addWidget(widget)
    assert widget.graphicsEffect() is None
    add_flash_animation(widget)
    assert widget.graphicsEffect() is not None
    assert hasattr(widget, '_flash_animation')
    qtbot.wait(350)
    assert widget.graphicsEffect() is None
    assert not hasattr(widget, '_flash_animation')


def test_qt_might_be_rich_text(qtbot):
    widget = QMainWindow()
    qtbot.addWidget(widget)
    assert qt_might_be_rich_text('<b>rich text</b>')
    assert not qt_might_be_rich_text('plain text')


def test_thread_proxy_guard(monkeypatch, qapp, single_threaded_executor):
    class X:
        a = 1

    monkeypatch.setenv('NAPARI_ENSURE_PLUGIN_MAIN_THREAD', 'True')

    x = X()
    x_proxy = PublicOnlyProxy(x)

    f = single_threaded_executor.submit(x.__setattr__, 'a', 2)
    f.result()
    assert x.a == 2

    f = single_threaded_executor.submit(x_proxy.__setattr__, 'a', 3)
    with pytest.raises(RuntimeError):
        f.result()
    assert x.a == 2


def test_get_color(qtbot):
    """Test the get_color utility function."""
    widget = QMainWindow()
    qtbot.addWidget(widget)

    with patch.object(QColorDialog, 'exec_') as mock:
        mock.return_value = QColorDialog.Accepted
        color = get_color(None, 'hex')
        assert isinstance(color, str), 'Expected string color'

    with patch.object(QColorDialog, 'exec_') as mock:
        mock.return_value = QColorDialog.Accepted
        color = get_color('#FF00FF', 'hex')
        assert isinstance(color, str), 'Expected string color'
        assert color == '#ff00ff', 'Expected color to be #FF00FF'

    with patch.object(QColorDialog, 'exec_') as mock:
        mock.return_value = QColorDialog.Accepted
        color = get_color(None, 'array')
        assert not isinstance(color, str), 'Expected array color'
        assert isinstance(color, np.ndarray), 'Expected numpy array color'

    with patch.object(QColorDialog, 'exec_') as mock:
        mock.return_value = QColorDialog.Accepted
        color = get_color(np.asarray([255, 0, 255]), 'array')
        assert not isinstance(color, str), 'Expected array color'
        assert isinstance(color, np.ndarray), 'Expected numpy array color'
        np.testing.assert_array_equal(color, np.asarray([1, 0, 1]))

    with patch.object(QColorDialog, 'exec_') as mock:
        mock.return_value = QColorDialog.Accepted
        color = get_color(None, 'qcolor')
        assert not isinstance(color, np.ndarray), 'Expected QColor color'
        assert isinstance(color, QColor), 'Expected QColor color'

    with patch.object(QColorDialog, 'exec_') as mock:
        mock.return_value = QColorDialog.Rejected
        color = get_color(None, 'qcolor')
        assert color is None, 'Expected None color'

    # close still open popup widgets
    for widget in QApplication.topLevelWidgets():
        if isinstance(widget, QColorDialog):
            widget.accept()
            widget.close()



import numpy as np
import pandas as pd
import pytest

from napari.layers import Tracks
from napari.layers.tracks._track_utils import TrackManager
from napari.utils._test_utils import (
    validate_all_params_in_docstring,
    validate_kwargs_sorted,
)

# def test_empty_tracks():
#     """Test instantiating Tracks layer without data."""
#     pts = Tracks()
#     assert pts.data.shape == (0, 4)


data_array_2dt = np.zeros((1, 4))
data_list_2dt = list(data_array_2dt)
dataframe_2dt = pd.DataFrame(
    data=data_array_2dt, columns=['track_id', 't', 'y', 'x']
)


@pytest.mark.parametrize(
    'data', [data_array_2dt, data_list_2dt, dataframe_2dt]
)
def test_tracks_layer_2dt_ndim(data):
    """Test instantiating Tracks layer, check 2D+t dimensionality."""
    layer = Tracks(data)
    assert layer.ndim == 3


data_array_3dt = np.zeros((1, 5))
data_list_3dt = list(data_array_3dt)
dataframe_3dt = pd.DataFrame(
    data=data_array_3dt, columns=['track_id', 't', 'z', 'y', 'x']
)


@pytest.mark.parametrize(
    'data', [data_array_3dt, data_list_3dt, dataframe_3dt]
)
def test_tracks_layer_3dt_ndim(data):
    """Test instantiating Tracks layer, check 3D+t dimensionality."""
    layer = Tracks(data)
    assert layer.ndim == 4


def test_track_layer_name():
    """Test track name."""
    data = np.zeros((1, 4))
    layer = Tracks(data, name='test_tracks')
    assert layer.name == 'test_tracks'


def test_track_layer_data():
    """Test data."""
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    layer = Tracks(data)
    np.testing.assert_array_equal(layer.data, data)


@pytest.mark.parametrize(
    'timestamps', [np.arange(100, 200), np.arange(100, 300, 2)]
)
def test_track_layer_data_nonzero_starting_time(timestamps):
    """Test data with sparse timestamps or not starting at zero."""
    data = np.zeros((100, 4))
    data[:, 1] = timestamps
    layer = Tracks(data)
    np.testing.assert_array_equal(layer.data, data)


def test_track_layer_data_flipped():
    """Test data flipped."""
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    data[:, 0] = np.arange(100)
    data = np.flip(data, axis=0)
    layer = Tracks(data)
    np.testing.assert_array_equal(layer.data, np.flip(data, axis=0))


properties_dict = {'time': np.arange(100)}
properties_df = pd.DataFrame(properties_dict)


@pytest.mark.parametrize('properties', [{}, properties_dict, properties_df])
def test_track_layer_properties(properties):
    """Test properties."""
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    layer = Tracks(data, properties=properties)
    for k, v in properties.items():
        np.testing.assert_equal(layer.properties[k], v)


@pytest.mark.parametrize('properties', [{}, properties_dict, properties_df])
def test_track_layer_properties_flipped(properties):
    """Test properties."""
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    data[:, 0] = np.arange(100)
    data = np.flip(data, axis=0)
    layer = Tracks(data, properties=properties)
    for k, v in properties.items():
        np.testing.assert_equal(layer.properties[k], np.flip(v))


@pytest.mark.filterwarnings('ignore:.*track_id.*:UserWarning')
def test_track_layer_colorby_nonexistent():
    """Test error handling for non-existent properties with color_by"""
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    non_existant_property = 'not_a_valid_key'
    assert non_existant_property not in properties_dict
    with pytest.raises(ValueError, match='not a valid property'):
        Tracks(
            data, properties=properties_dict, color_by=non_existant_property
        )


@pytest.mark.filterwarnings('ignore:.*track_id.*:UserWarning')
def test_track_layer_properties_changed_colorby():
    """Test behaviour when changes to properties invalidate current color_by"""
    properties_dict_1 = {'time': np.arange(100), 'prop1': np.arange(100)}
    properties_dict_2 = {'time': np.arange(100), 'prop2': np.arange(100)}
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    layer = Tracks(data, properties=properties_dict_1, color_by='prop1')
    # test warning is raised
    with pytest.warns(UserWarning):
        layer.properties = properties_dict_2
    # test default fallback
    assert layer.color_by == 'track_id'


def test_track_layer_graph():
    """Test track layer graph."""
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    data[50:, 0] = 1
    graph = {1: [0]}
    layer = Tracks(data, graph=graph)
    assert layer.graph == graph


def test_track_layer_reset_data():
    """Test changing data once layer is instantiated."""
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    data[50:, 0] = 1
    properties = {'time': data[:, 1]}
    graph = {1: [0]}
    layer = Tracks(data, graph=graph, properties=properties)
    cropped_data = data[:10, :]
    layer.data = cropped_data
    np.testing.assert_array_equal(layer.data, cropped_data)
    assert layer.graph == {}


def test_malformed_id():
    """Test for malformed track ID."""
    data = np.random.random((100, 4))
    data[:, 1] = np.arange(100)
    with pytest.raises(ValueError, match='must be an integer'):
        Tracks(data)


def test_malformed_graph():
    """Test for malformed graph."""
    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    data[50:, 0] = 1
    graph = {1: [0], 2: [33]}
    with pytest.raises(ValueError, match='node 2 not found'):
        Tracks(data, graph=graph)


def test_tracks_float_time_index():
    """Test Tracks layer instantiation with floating point time values"""
    coords = np.random.normal(loc=50, size=(100, 2))
    time = np.random.normal(loc=50, size=(100, 1))
    track_id = np.zeros((100, 1))
    track_id[50:] = 1
    data = np.concatenate((track_id, time, coords), axis=1)
    Tracks(data)


def test_tracks_length_change():
    """Test changing length properties of tracks"""
    track_length = 1000
    data = np.zeros((track_length, 4))
    layer = Tracks(data)
    layer.tail_length = track_length
    assert layer.tail_length == track_length
    assert layer._max_length == track_length

    layer = Tracks(data)
    layer.head_length = track_length
    assert layer.head_length == track_length
    assert layer._max_length == track_length


def test_fast_points_lookup() -> None:
    # creates sorted points
    time_points = np.asarray([0, 1, 3, 5, 10])
    repeats = np.asarray([3, 4, 6, 3, 5])
    sorted_time = np.repeat(time_points, repeats)
    end = np.cumsum(repeats)
    start = np.insert(end[:-1], 0, 0)

    # compute lookup
    points_lookup = TrackManager._fast_points_lookup(sorted_time)

    assert len(time_points) == len(points_lookup)
    total_length = 0
    for s, e, t, r in zip(start, end, time_points, repeats, strict=False):
        assert points_lookup[t].start == s
        assert points_lookup[t].stop == e
        assert points_lookup[t].stop - points_lookup[t].start == r
        unique_time = sorted_time[points_lookup[t]]
        np.testing.assert_array_equal(unique_time[0], unique_time)
        total_length += len(unique_time)

    assert total_length == len(sorted_time)


def test_single_time_tracks() -> None:
    """Edge case where all tracks belong to a single time"""

    # track_id, t, y, x
    tracks = [[0, 5, 2, 3], [1, 5, 3, 4], [2, 5, 4, 5]]
    layer = Tracks(tracks)

    np.testing.assert_array_equal(layer.data, tracks)


def test_track_ids_ordering() -> None:
    """Check if tracks ids are correctly set to features when given not-sorted tracks."""
    # track_id, t, y, x
    unsorted_data = np.asarray(
        [[1, 1, 0, 0], [0, 1, 0, 0], [2, 0, 0, 0], [0, 0, 0, 0], [1, 0, 0, 0]]
    )
    sorted_track_ids = [0, 0, 1, 1, 2]  # track_ids after sorting

    layer = Tracks(unsorted_data)
    np.testing.assert_array_equal(sorted_track_ids, layer.features['track_id'])


def test_changing_data_inplace() -> None:
    """Test if layer can be refreshed after changing data in place."""

    data = np.ones((100, 4))
    data[:, 1] = np.arange(100)

    layer = Tracks(data)

    # Change data in place
    # coordinates
    layer.data[50:, -1] = 2
    layer.refresh()

    # time
    layer.data[50:, 1] = np.arange(100, 150)
    layer.refresh()

    # track_id
    layer.data[50:, 0] = 2
    layer.refresh()


def test_track_connex_validity() -> None:
    """Test if track_connex is valid (i.e if the value False appears as many
    times as there are tracks."""

    data = np.zeros((11, 4))

    # Track ids
    data[:-1, 0] = np.repeat(np.arange(1, 6), 2)
    # create edge case where a track has length one
    data[-1, 0] = 6

    # Time
    data[:-1, 1] = np.array([0, 1] * 5)
    data[-1, 1] = 0

    layer = Tracks(data)

    # number of tracks
    n_tracks = 6

    # the number of 'False' in the track_connex array should be equal to the number of tracks
    assert np.sum(~layer._manager.track_connex) == n_tracks


def test_track_coloring() -> None:
    """Test if the track colors are correctly set."""

    data = np.zeros((100, 4))
    data[:, 1] = np.arange(100)
    layer = Tracks(data)

    colors = np.random.random(size=(100, 4))
    layer.track_colors = colors

    assert np.array_equal(layer._track_colors, colors)


def test_docstring():
    validate_all_params_in_docstring(Tracks)
    validate_kwargs_sorted(Tracks)



import typing
import warnings
from collections import deque
from collections.abc import Callable, Sequence
from contextlib import contextmanager
from typing import (
    Any,
    ClassVar,
)

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import ndimage as ndi
from skimage.draw import polygon2mask

from napari.layers._data_protocols import LayerDataProtocol
from napari.layers._multiscale_data import MultiScaleData
from napari.layers._scalar_field.scalar_field import ScalarFieldBase
from napari.layers.base import Layer, no_op
from napari.layers.base._base_mouse_bindings import (
    highlight_box_handles,
    transform_with_box,
)
from napari.layers.image._image_utils import guess_multiscale
from napari.layers.image._slice import _ImageSliceResponse
from napari.layers.labels._labels_constants import (
    IsoCategoricalGradientMode,
    LabelColorMode,
    LabelsRendering,
    Mode,
)
from napari.layers.labels._labels_mouse_bindings import (
    BrushSizeOnMouseMove,
    draw,
    pick,
)
from napari.layers.labels._labels_utils import (
    expand_slice,
    get_contours,
    indices_in_shape,
    interpolate_coordinates,
    sphere_indices,
)
from napari.layers.utils.layer_utils import _FeatureTable
from napari.utils._dtype import normalize_dtype, vispy_texture_dtype
from napari.utils._indexing import elements_in_slice, index_in_slice
from napari.utils.colormaps import (
    direct_colormap,
    label_colormap,
)
from napari.utils.colormaps.colormap import (
    CyclicLabelColormap,
    LabelColormapBase,
    _normalize_label_colormap,
)
from napari.utils.colormaps.colormap_utils import shuffle_and_extend_colormap
from napari.utils.events import EmitterGroup, Event
from napari.utils.events.custom_types import Array
from napari.utils.misc import StringEnum, _is_array_type
from napari.utils.naming import magic_name
from napari.utils.translations import trans

__all__ = ('Labels',)


class Labels(ScalarFieldBase):
    """Labels (or segmentation) layer.

    An image-like layer where every pixel contains an integer ID
    corresponding to the region it belongs to.

    Parameters
    ----------
    data : array or list of array
        Labels data as an array or multiscale. Must be integer type or bools.
        Please note multiscale rendering is only supported in 2D. In 3D, only
        the lowest resolution scale is displayed.
    affine : n-D array or napari.utils.transforms.Affine
        (N+1, N+1) affine transformation matrix in homogeneous coordinates.
        The first (N, N) entries correspond to a linear transform and
        the final column is a length N translation vector and a 1 or a napari
        `Affine` transform object. Applied as an extra transform on top of the
        provided scale, rotate, and shear values.
    axis_labels : tuple of str, optional
        Dimension names of the layer data.
        If not provided, axis_labels will be set to (..., 'axis -2', 'axis -1').
    blending : str
        One of a list of preset blending modes that determines how RGB and
        alpha values of the layer visual get mixed. Allowed values are
        {'opaque', 'translucent', and 'additive'}.
    cache : bool
        Whether slices of out-of-core datasets should be cached upon retrieval.
        Currently, this only applies to dask arrays.
    colormap : CyclicLabelColormap or DirectLabelColormap or None
        Colormap to use for the labels. If None, a random colormap will be
        used.
    depiction : str
        3D Depiction mode. Must be one of {'volume', 'plane'}.
        The default value is 'volume'.
    experimental_clipping_planes : list of dicts, list of ClippingPlane, or ClippingPlaneList
        Each dict defines a clipping plane in 3D in data coordinates.
        Valid dictionary keys are {'position', 'normal', and 'enabled'}.
        Values on the negative side of the normal are discarded if the plane is enabled.
    features : dict[str, array-like] or DataFrame
        Features table where each row corresponds to a label and each column
        is a feature. The first row corresponds to the background label.
    iso_gradient_mode : str
        Method for calulating the gradient (used to get the surface normal) in the
        'iso_categorical' rendering mode. Must be one of {'fast', 'smooth'}.
        'fast' uses a simple finite difference gradient in x, y, and z. 'smooth' uses an
        isotropic Sobel gradient, which is smoother but more computationally expensive.
        The default value is 'fast'.
    metadata : dict
        Layer metadata.
    multiscale : bool
        Whether the data is a multiscale image or not. Multiscale data is
        represented by a list of array like image data. If not specified by
        the user and if the data is a list of arrays that decrease in shape
        then it will be taken to be multiscale. The first image in the list
        should be the largest. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    name : str
        Name of the layer.
    opacity : float
        Opacity of the layer visual, between 0.0 and 1.0.
    plane : dict or SlicingPlane
        Properties defining plane rendering in 3D. Properties are defined in
        data coordinates. Valid dictionary keys are
        {'position', 'normal', 'thickness', and 'enabled'}.
    projection_mode : str
        How data outside the viewed dimensions but inside the thick Dims slice will
        be projected onto the viewed dimensions
    properties : dict {str: array (N,)} or DataFrame
        Properties for each label. Each property should be an array of length
        N, where N is the number of labels, and the first property corresponds
        to background.
    rendering : str
        3D Rendering mode used by vispy. Must be one {'translucent', 'iso_categorical'}.
        'translucent' renders without lighting. 'iso_categorical' uses isosurface
        rendering to calculate lighting effects on labeled surfaces.
        The default value is 'iso_categorical'.
    rotate : float, 3-tuple of float, or n-D array.
        If a float convert into a 2D rotation matrix using that value as an
        angle. If 3-tuple convert into a 3D rotation matrix, using a yaw,
        pitch, roll convention. Otherwise assume an nD rotation. Angles are
        assumed to be in degrees. They can be converted from radians with
        np.degrees if needed.
    scale : tuple of float
        Scale factors for the layer.
    shear : 1-D array or n-D array
        Either a vector of upper triangular values, or an nD shear matrix with
        ones along the main diagonal.
    translate : tuple of float
        Translation values for the layer.
    units : tuple of str or pint.Unit, optional
        Units of the layer data in world coordinates.
        If not provided, the default units are assumed to be pixels.
    visible : bool
        Whether the layer visual is currently being displayed.

    Attributes
    ----------
    data : array or list of array
        Integer label data as an array or multiscale. Can be N dimensional.
        Every pixel contains an integer ID corresponding to the region it
        belongs to. The label 0 is rendered as transparent. Please note
        multiscale rendering is only supported in 2D. In 3D, only
        the lowest resolution scale is displayed.
    axis_labels : tuple of str
        Dimension names of the layer data.
    multiscale : bool
        Whether the data is a multiscale image or not. Multiscale data is
        represented by a list of array like image data. The first image in the
        list should be the largest. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    metadata : dict
        Labels metadata.
    num_colors : int
        Number of unique colors to use in colormap. DEPRECATED: set
        ``colormap`` directly, using `napari.utils.colormaps.label_colormap`.
    features : Dataframe-like
        Features table where each row corresponds to a label and each column
        is a feature. The first row corresponds to the background label.
    properties : dict {str: array (N,)}, DataFrame
        Properties for each label. Each property should be an array of length
        N, where N is the number of labels, and the first property corresponds
        to background.
    color : dict of int to str or array
        Custom label to color mapping. Values must be valid color names or RGBA
        arrays. While there is no limit to the number of custom labels, the
        the layer will render incorrectly if they map to more than 1024 distinct
        colors. DEPRECATED: set ``colormap`` directly, using
        `napari.utils.colormaps.DirectLabelColormap`.
    seed : float
        Seed for colormap random generator. DEPRECATED: set ``colormap``
        directly, using `napari.utils.colormaps.label_colormap`.
    opacity : float
        Opacity of the labels, must be between 0 and 1.
    contiguous : bool
        If `True`, the fill bucket changes only connected pixels of same label.
    n_edit_dimensions : int
        The number of dimensions across which labels will be edited.
    contour : int
        If greater than 0, displays contours of labels instead of shaded regions
        with a thickness equal to its value. Must be >= 0.
    brush_size : float
        Size of the paint brush in data coordinates.
    iso_gradient_mode : str
        Method for calulating the gradient (used to get the surface normal) in the
        'iso_categorical' rendering mode. Must be one of {'fast', 'smooth'}.
        'fast' uses a simple finite difference gradient in x, y, and z. 'smooth' uses an
        isotropic Sobel gradient, which is smoother but more computationally expensive.
    selected_label : int
        Index of selected label. Can be greater than the current maximum label.
    mode : str
        Interactive mode. The normal, default mode is PAN_ZOOM, which
        allows for normal interactivity with the canvas.

        In PICK mode the cursor functions like a color picker, setting the
        clicked on label to be the current label. If the background is picked it
        will select the background label `0`.

        In PAINT mode the cursor functions like a paint brush changing any
        pixels it brushes over to the current label. If the background label
        `0` is selected than any pixels will be changed to background and this
        tool functions like an eraser. The size and shape of the cursor can be
        adjusted in the properties widget.

        In FILL mode the cursor functions like a fill bucket replacing pixels
        of the label clicked on with the current label. It can either replace
        all pixels of that label or just those that are contiguous with the
        clicked on pixel. If the background label `0` is selected than any
        pixels will be changed to background and this tool functions like an
        eraser.

        In ERASE mode the cursor functions similarly to PAINT mode, but to
        paint with background label, which effectively removes the label.
    plane : SlicingPlane
        Properties defining plane rendering in 3D.
    experimental_clipping_planes : ClippingPlaneList
        Clipping planes defined in data coordinates, used to clip the volume.
    units: tuple of pint.Unit
        Units of the layer data in world coordinates.

    Notes
    -----
    _selected_color : 4-tuple or None
        RGBA tuple of the color of the selected label, or None if the
        background label `0` is selected.
    """

    events: EmitterGroup
    _colormap: LabelColormapBase

    _modeclass = Mode

    _drag_modes: ClassVar[dict[Mode, Callable[['Labels', Event], None]]] = {  # type: ignore[assignment]
        Mode.PAN_ZOOM: no_op,
        Mode.TRANSFORM: transform_with_box,
        Mode.PICK: pick,
        Mode.PAINT: draw,
        Mode.FILL: draw,
        Mode.ERASE: draw,
        Mode.POLYGON: no_op,  # the overlay handles mouse events in this mode
    }

    brush_size_on_mouse_move = BrushSizeOnMouseMove(min_brush_size=1)

    _move_modes: ClassVar[
        dict[StringEnum, Callable[['Labels', Event], None]]
    ] = {  # type: ignore[assignment]
        Mode.PAN_ZOOM: no_op,
        Mode.TRANSFORM: highlight_box_handles,
        Mode.PICK: no_op,
        Mode.PAINT: brush_size_on_mouse_move,
        Mode.FILL: no_op,
        Mode.ERASE: brush_size_on_mouse_move,
        Mode.POLYGON: no_op,  # the overlay handles mouse events in this mode
    }

    _cursor_modes: ClassVar[dict[Mode, str]] = {  # type: ignore[assignment]
        Mode.PAN_ZOOM: 'standard',
        Mode.TRANSFORM: 'standard',
        Mode.PICK: 'cross',
        Mode.PAINT: 'circle',
        Mode.FILL: 'cross',
        Mode.ERASE: 'circle',
        Mode.POLYGON: 'cross',
    }

    _history_limit = 100

    def __init__(
        self,
        data,
        *,
        affine=None,
        axis_labels=None,
        blending='translucent',
        cache=True,
        colormap=None,
        depiction='volume',
        experimental_clipping_planes=None,
        features=None,
        iso_gradient_mode=IsoCategoricalGradientMode.FAST.value,
        metadata=None,
        multiscale=None,
        name=None,
        opacity=0.7,
        plane=None,
        projection_mode='none',
        properties=None,
        rendering='iso_categorical',
        rotate=None,
        scale=None,
        shear=None,
        translate=None,
        units=None,
        visible=True,
    ) -> None:
        if name is None and data is not None:
            name = magic_name(data)

        self._seed = 0.5
        # We use 50 colors (49 + transparency) by default for historical
        # consistency. This may change in future versions.
        self._random_colormap = label_colormap(
            49, self._seed, background_value=0
        )
        self._original_random_colormap = self._random_colormap
        self._direct_colormap = direct_colormap(
            {0: 'transparent', None: 'black'}
        )
        self._colormap = self._random_colormap
        self._color_mode = LabelColorMode.AUTO
        self._show_selected_label = False
        self._contour = 0

        data = self._ensure_int_labels(data)

        super().__init__(
            data,
            affine=affine,
            axis_labels=axis_labels,
            blending=blending,
            cache=cache,
            depiction=depiction,
            experimental_clipping_planes=experimental_clipping_planes,
            rendering=rendering,
            metadata=metadata,
            multiscale=multiscale,
            name=name,
            scale=scale,
            shear=shear,
            plane=plane,
            opacity=opacity,
            projection_mode=projection_mode,
            rotate=rotate,
            translate=translate,
            units=units,
            visible=visible,
        )

        self.events.add(
            brush_shape=Event,
            brush_size=Event,
            colormap=Event,
            contiguous=Event,
            contour=Event,
            features=Event,
            iso_gradient_mode=Event,
            labels_update=Event,
            n_edit_dimensions=Event,
            paint=Event,
            preserve_labels=Event,
            properties=Event,
            selected_label=Event,
            show_selected_label=Event,
        )

        from napari.components.overlays.labels_polygon import (
            LabelsPolygonOverlay,
        )

        self._overlays.update({'polygon': LabelsPolygonOverlay()})

        self._feature_table = _FeatureTable.from_layer(
            features=features, properties=properties
        )
        self._label_index = self._make_label_index()

        self._n_edit_dimensions = 2
        self._contiguous = True
        self._brush_size = 10

        self._iso_gradient_mode = IsoCategoricalGradientMode(iso_gradient_mode)

        self._selected_label = 1
        self.colormap.selection = self._selected_label
        self.colormap.use_selection = self._show_selected_label
        self._prev_selected_label = None
        self._selected_color = self.get_color(self._selected_label)
        self._updated_slice = None
        if colormap is not None:
            self._set_colormap(colormap)

        self._status = self.mode
        self._preserve_labels = False

    def _post_init(self):
        self._reset_history()
        # Trigger generation of view slice and thumbnail
        self.refresh()
        self._reset_editable()

    @property
    def rendering(self):
        """Return current rendering mode.

        Selects a preset rendering mode in vispy that determines how
        lablels are displayed.  Options include:

        * ``translucent``: voxel colors are blended along the view ray until
          the result is opaque.
        * ``iso_categorical``: isosurface for categorical data.
          Cast a ray until a non-background value is encountered. At that
          location, lighning calculations are performed to give the visual
          appearance of a surface.

        Returns
        -------
        str
            The current rendering mode
        """
        return str(self._rendering)

    @rendering.setter
    def rendering(self, rendering):
        self._rendering = LabelsRendering(rendering)
        self.events.rendering()

    @property
    def iso_gradient_mode(self) -> str:
        """Return current gradient mode for isosurface rendering.

        Selects the finite-difference gradient method for the isosurface shader. Options include:
            * ``fast``: use a simple finite difference gradient along each axis
            * ``smooth``: use an isotropic Sobel gradient, smoother but more
              computationally expensive

        Returns
        -------
        str
            The current gradient mode
        """
        return str(self._iso_gradient_mode)

    @iso_gradient_mode.setter
    def iso_gradient_mode(self, value: IsoCategoricalGradientMode | str):
        self._iso_gradient_mode = IsoCategoricalGradientMode(value)
        self.events.iso_gradient_mode()

    @property
    def contiguous(self):
        """bool: fill bucket changes only connected pixels of same label."""
        return self._contiguous

    @contiguous.setter
    def contiguous(self, contiguous):
        self._contiguous = contiguous
        self.events.contiguous()

    @property
    def n_edit_dimensions(self):
        return self._n_edit_dimensions

    @n_edit_dimensions.setter
    def n_edit_dimensions(self, n_edit_dimensions):
        self._n_edit_dimensions = n_edit_dimensions
        self.events.n_edit_dimensions()

    @property
    def contour(self) -> int:
        """int: displays contours of labels instead of shaded regions."""
        return self._contour

    @contour.setter
    def contour(self, contour: int) -> None:
        if contour < 0:
            raise ValueError('contour value must be >= 0')
        self._contour = int(contour)
        self.events.contour()
        self.refresh(extent=False)

    @property
    def brush_size(self):
        """float: Size of the paint in world coordinates."""
        return self._brush_size

    @brush_size.setter
    def brush_size(self, brush_size):
        self._brush_size = int(brush_size)
        self.cursor_size = self._calculate_cursor_size()
        self.events.brush_size()

    def _calculate_cursor_size(self):
        # Convert from brush size in data coordinates to
        # cursor size in world coordinates
        scale = self._data_to_world.scale
        min_scale = np.min(
            [abs(scale[d]) for d in self._slice_input.displayed]
        )
        return abs(self.brush_size * min_scale)

    def new_colormap(self, seed: int | None = None):
        if seed is None:
            seed = np.random.default_rng().integers(2**32 - 1)

        orig = self._original_random_colormap
        self.colormap = shuffle_and_extend_colormap(
            self._original_random_colormap, seed
        )
        self._original_random_colormap = orig

    @property
    def colormap(self) -> LabelColormapBase:
        return self._colormap

    @colormap.setter
    def colormap(self, colormap: LabelColormapBase):
        self._set_colormap(colormap)

    def _set_colormap(self, colormap):
        colormap = _normalize_label_colormap(colormap)
        if isinstance(colormap, CyclicLabelColormap):
            self._random_colormap = colormap
            self._original_random_colormap = colormap
            self._colormap = self._random_colormap
            color_mode = LabelColorMode.AUTO
        else:
            self._direct_colormap = colormap
            # `self._direct_colormap.color_dict` may contain just the default None and background label
            # colors, in which case we need to be in AUTO color mode. Otherwise,
            # `self._direct_colormap.color_dict` contains colors for all labels, and we should be in DIRECT
            # mode.

            # For more information
            # - https://github.com/napari/napari/issues/2479
            # - https://github.com/napari/napari/issues/2953
            if self._is_default_colors(self._direct_colormap.color_dict):
                color_mode = LabelColorMode.AUTO
                self._colormap = self._random_colormap
            else:
                color_mode = LabelColorMode.DIRECT
                self._colormap = self._direct_colormap
        self._cached_labels = None  # invalidate the cached color mapping
        self._selected_color = self.get_color(self.selected_label)
        self._color_mode = color_mode
        self.events.colormap()  # Will update the LabelVispyColormap shader
        self.events.selected_label()
        self.refresh(extent=False)

    @property
    def data(self) -> LayerDataProtocol | MultiScaleData:
        """array: Image data."""
        return self._data

    @data.setter
    def data(self, data: LayerDataProtocol | MultiScaleData):
        data = self._ensure_int_labels(data)
        self._data = data
        self._ndim = len(self._data.shape)
        self._update_dims()
        self.events.data(value=self.data)
        self._reset_editable()

    @property
    def features(self):
        """Dataframe-like features table.

        It is an implementation detail that this is a `pandas.DataFrame`. In the future,
        we will target the currently-in-development Data API dataframe protocol [1]_.
        This will enable us to use alternate libraries such as xarray or cuDF for
        additional features without breaking existing usage of this.

        If you need to specifically rely on the pandas API, please coerce this to a
        `pandas.DataFrame` using `features_to_pandas_dataframe`.

        References
        ----------
        .. [1] https://data-apis.org/dataframe-protocol/latest/API.html
        """
        return self._feature_table.values

    @features.setter
    def features(
        self,
        features: dict[str, np.ndarray] | pd.DataFrame,
    ) -> None:
        self._feature_table.set_values(features)
        self._label_index = self._make_label_index()
        self.events.properties()
        self.events.features()

    @property
    def properties(self) -> dict[str, np.ndarray]:
        """dict {str: array (N,)}, DataFrame: Properties for each label."""
        return self._feature_table.properties()

    @properties.setter
    def properties(self, properties: dict[str, Array]):
        self.features = properties

    def _make_label_index(self) -> dict[int, int]:
        features = self._feature_table.values
        label_index = {}
        if 'index' in features:
            label_index = {i: k for k, i in enumerate(features['index'])}
        elif features.shape[1] > 0:
            label_index = {i: i for i in range(features.shape[0])}
        return label_index

    def _is_default_colors(self, color):
        """Returns True if color contains only default colors, otherwise False.

        Default colors are black for `None` and transparent for
        `self.colormap.background_value`.

        Parameters
        ----------
        color : Dict
            Dictionary of label value to color array

        Returns
        -------
        bool
            True if color contains only default colors, otherwise False.
        """
        return (
            {None, self.colormap.background_value} == set(color.keys())
            and np.allclose(color[None], [0, 0, 0, 1])
            and np.allclose(
                color[self.colormap.background_value], [0, 0, 0, 0]
            )
        )

    def _ensure_int_labels(self, data):
        """Ensure data is integer by converting from bool if required, raising an error otherwise."""
        looks_multiscale, data = guess_multiscale(data)
        if not looks_multiscale:
            data = [data]
        int_data = []
        for data_level in data:
            # normalize_dtype turns e.g. tensorstore or torch dtypes into
            # numpy dtypes
            if np.issubdtype(normalize_dtype(data_level.dtype), np.floating):
                raise TypeError(
                    trans._(
                        'Only integer types are supported for Labels layers, but data contains {data_level_type}.',
                        data_level_type=data_level.dtype,
                    )
                )
            if data_level.dtype == bool:
                int_data.append(data_level.view(np.uint8))
            else:
                int_data.append(data_level)
        data = int_data
        if not looks_multiscale:
            data = data[0]
        return data

    def _get_state(self) -> dict[str, Any]:
        """Get dictionary of layer state.

        Returns
        -------
        state : dict of str to Any
            Dictionary of layer state.
        """
        state = self._get_base_state()
        state.update(
            {
                'multiscale': self.multiscale,
                'properties': self.properties,
                'rendering': self.rendering,
                'iso_gradient_mode': self.iso_gradient_mode,
                'depiction': self.depiction,
                'plane': self.plane.dict(),
                'experimental_clipping_planes': [
                    plane.dict() for plane in self.experimental_clipping_planes
                ],
                'data': self.data,
                'features': self.features,
                'colormap': self.colormap,
            }
        )
        return state

    @property
    def selected_label(self):
        """int: Index of selected label."""
        return self._selected_label

    @selected_label.setter
    def selected_label(self, selected_label):
        if selected_label == self.selected_label:
            return
        # when setting the label to the background, store the previous
        # otherwise, clear it
        if selected_label == self.colormap.background_value:
            self._prev_selected_label = self.selected_label
        else:
            self._prev_selected_label = None
        self.colormap.selection = selected_label
        self._selected_label = selected_label
        self._selected_color = self.get_color(selected_label)

        self.events.selected_label()

        if self.show_selected_label:
            self.refresh(extent=False)

    def swap_selected_and_background_labels(self):
        """Swap between the selected label and the background label."""
        if self.selected_label != self.colormap.background_value:
            self.selected_label = self.colormap.background_value
        else:
            self.selected_label = self._prev_selected_label

    @property
    def show_selected_label(self):
        """Whether to filter displayed labels to only the selected label or not"""
        return self._show_selected_label

    @show_selected_label.setter
    def show_selected_label(self, show_selected):
        self._show_selected_label = show_selected
        self.colormap.use_selection = show_selected
        self.colormap.selection = self.selected_label
        self.events.show_selected_label(show_selected_label=show_selected)
        self.refresh(extent=False)

    # Only overriding to change the docstring
    @property
    def mode(self):
        """MODE: Interactive mode. The normal, default mode is PAN_ZOOM, which
        allows for normal interactivity with the canvas.

        In PICK mode the cursor functions like a color picker, setting the
        clicked on label to be the current label. If the background is picked it
        will select the background label `0`.

        In PAINT mode the cursor functions like a paint brush changing any
        pixels it brushes over to the current label. If the background label
        `0` is selected than any pixels will be changed to background and this
        tool functions like an eraser. The size and shape of the cursor can be
        adjusted in the properties widget.

        In FILL mode the cursor functions like a fill bucket replacing pixels
        of the label clicked on with the current label. It can either replace
        all pixels of that label or just those that are contiguous with the
        clicked on pixel. If the background label `0` is selected than any
        pixels will be changed to background and this tool functions like an
        eraser.

        In ERASE mode the cursor functions similarly to PAINT mode, but to
        paint with background label, which effectively removes the label.
        """
        return Layer.mode.fget(self)

    # Only overriding to change the docstring of the setter above
    @mode.setter
    def mode(self, mode):
        Layer.mode.fset(self, mode)

    def _mode_setter_helper(self, mode):
        mode = super()._mode_setter_helper(mode)
        if mode == self._mode:
            return mode

        self._overlays['polygon'].enabled = mode == Mode.POLYGON
        if mode in {Mode.PAINT, Mode.ERASE}:
            self.cursor_size = self._calculate_cursor_size()

        return mode

    @property
    def preserve_labels(self):
        """Defines if painting should preserve existing labels.

        Default to false to allow paint on existing labels. When
        set to true, existing labels will be preserved during painting.
        """
        return self._preserve_labels

    @preserve_labels.setter
    def preserve_labels(self, preserve_labels: bool):
        self._preserve_labels = preserve_labels
        self.events.preserve_labels(preserve_labels=preserve_labels)

    def _reset_editable(self) -> None:
        self.editable = not self.multiscale

    def _on_editable_changed(self) -> None:
        if not self.editable:
            self.mode = Mode.PAN_ZOOM
            self._reset_history()

    @staticmethod
    def _to_vispy_texture_dtype(data):
        """Convert data to a dtype that can be used as a VisPy texture.

        Labels layers allow all integer dtypes for data, but only a subset
        are supported by VisPy textures. For now, we convert all data to
        float32 as it can represent all input values (though not losslessly,
        see https://github.com/napari/napari/issues/6084).
        """
        return vispy_texture_dtype(data)

    def _update_slice_response(self, response: _ImageSliceResponse) -> None:
        """Override to convert raw slice data to displayed label colors."""
        response = response.to_displayed(self._raw_to_displayed)
        super()._update_slice_response(response)

    def _partial_labels_refresh(self):
        """Prepares and displays only an updated part of the labels."""

        if self._updated_slice is None or not self.loaded:
            return

        dims_displayed = self._slice_input.displayed
        raw_displayed = self._slice.image.raw

        # Keep only the dimensions that correspond to the current view
        updated_slice = tuple(
            self._updated_slice[index] for index in dims_displayed
        )

        offset = [axis_slice.start for axis_slice in updated_slice]

        if self.contour > 0:
            colors_sliced = self._raw_to_displayed(
                raw_displayed, data_slice=updated_slice
            )
        else:
            colors_sliced = self._slice.image.view[updated_slice]
        # The next line is needed to make the following tests pass in
        # napari/_vispy/_tests/:
        # - test_vispy_labels_layer.py::test_labels_painting
        # - test_vispy_labels_layer.py::test_labels_fill_slice
        # See https://github.com/napari/napari/pull/6112/files#r1291613760
        # and https://github.com/napari/napari/issues/6185
        self._slice.image.view[updated_slice] = colors_sliced

        self.events.labels_update(data=colors_sliced, offset=offset)
        self._updated_slice = None

    def _calculate_contour(
        self, labels: np.ndarray, data_slice: tuple[slice, ...]
    ) -> np.ndarray | None:
        """Calculate the contour of a given label array within the specified data slice.

        Parameters
        ----------
        labels : np.ndarray
            The label array.
        data_slice : Tuple[slice, ...]
            The slice of the label array on which to calculate the contour.

        Returns
        -------
        Optional[np.ndarray]
            The calculated contour as a boolean mask array.
            Returns None if the contour parameter is less than 1,
            or if the label array has more than 2 dimensions.
        """
        if self.contour < 1:
            return None
        if labels.ndim > 2:
            warnings.warn(
                trans._(
                    'Contours are not displayed during 3D rendering',
                    deferred=True,
                )
            )
            return None

        expanded_slice = expand_slice(data_slice, labels.shape, 1)
        sliced_labels = get_contours(
            labels[expanded_slice],
            self.contour,
            self.colormap.background_value,
        )

        # Remove the latest one-pixel border from the result
        delta_slice = tuple(
            slice(s1.start - s2.start, s1.stop - s2.start)
            for s1, s2 in zip(data_slice, expanded_slice, strict=False)
        )
        return sliced_labels[delta_slice]

    def _raw_to_displayed(
        self, raw, data_slice: tuple[slice, ...] | None = None
    ) -> np.ndarray:
        """Determine displayed image from a saved raw image and a saved seed.

        This function ensures that the 0 label gets mapped to the 0 displayed
        pixel.

        Parameters
        ----------
        raw : array or int
            Raw integer input image.

        data_slice : numpy array slice
            Slice that specifies the portion of the input image that
            should be computed and displayed.
            If None, the whole input image will be processed.

        Returns
        -------
        mapped_labels : array
            Encoded colors mapped between 0 and 1 to be displayed.
        """

        if data_slice is None:
            data_slice = tuple(slice(0, size) for size in raw.shape)

        labels = raw  # for readability

        sliced_labels = self._calculate_contour(labels, data_slice)

        # lookup function -> self._as_type

        if sliced_labels is None:
            sliced_labels = labels[data_slice]

        return self.colormap._data_to_texture(sliced_labels)

    def _update_thumbnail(self):
        """Update the thumbnail with current data and colormap.

        This is overridden from _ImageBase because we don't need to do things
        like adjusting gamma or changing the data based on the contrast
        limits.
        """
        if not self.loaded:
            # ASYNC_TODO: Do not compute the thumbnail until we are loaded.
            # Is there a nicer way to prevent this from getting called?
            return

        image = self._slice.thumbnail.raw
        if self._slice_input.ndisplay == 3 and self.ndim > 2:
            # we are only using the current slice so `image` will never be
            # bigger than 3. If we are in this clause, it is exactly 3, so we
            # use max projection. For labels, ideally we would use "first
            # nonzero projection", but we leave that for a future PR. (TODO)
            image = np.max(image, axis=0)
        imshape = np.array(image.shape[:2])
        thumbshape = np.array(self._thumbnail_shape[:2])

        raw_zoom_factor = np.min(thumbshape / imshape)
        new_shape = np.clip(
            raw_zoom_factor * imshape, a_min=1, a_max=thumbshape
        )
        zoom_factor = tuple(new_shape / imshape)

        downsampled = ndi.zoom(image, zoom_factor, prefilter=False, order=0)
        color_array = self.colormap.map(downsampled)
        color_array[..., 3] *= self.opacity

        self.thumbnail = color_array

    def get_color(self, label):
        """Return the color corresponding to a specific label."""
        if label == self.colormap.background_value:
            col = None
        elif label is None or (
            self.show_selected_label and label != self.selected_label
        ):
            col = self.colormap.map(self.colormap.background_value)
        else:
            col = self.colormap.map(label)
        return col

    def _reset_history(self, event=None):
        self._undo_history = deque(maxlen=self._history_limit)
        self._redo_history = deque(maxlen=self._history_limit)
        self._staged_history = []
        self._block_history = False

    @contextmanager
    def block_history(self):
        """Context manager to group history-editing operations together.

        While in the context, history atoms are grouped together into a
        "staged" history. When exiting the context, that staged history is
        committed to the undo history queue, and an event is emitted
        containing the change.
        """
        prev = self._block_history
        self._block_history = True
        try:
            yield
            self._commit_staged_history()
        finally:
            self._block_history = prev

    def _commit_staged_history(self):
        """Save staged history to undo history and clear it."""
        if self._staged_history:
            self._append_to_undo_history(self._staged_history)
            self._staged_history = []

    def _append_to_undo_history(self, item):
        """Append item to history and emit paint event.

        Parameters
        ----------
        item : List[Tuple[ndarray, ndarray, int]]
            list of history atoms to append to undo history.
        """
        self._undo_history.append(item)
        self.events.paint(value=item)

    def _save_history(self, value):
        """Save a history "atom" to the undo history.

        A history "atom" is a single change operation to the array. A history
        *item* is a collection of atoms that were applied together to make a
        single change. For example, when dragging and painting, at each mouse
        callback we create a history "atom", but we save all those atoms in
        a single history item, since we would want to undo one drag in one
        undo operation.

        Parameters
        ----------
        value : 3-tuple of arrays
            The value is a 3-tuple containing:

            - a numpy multi-index, pointing to the array elements that were
              changed
            - the values corresponding to those elements before the change
            - the value(s) after the change
        """
        self._redo_history.clear()
        if self._block_history:
            self._staged_history.append(value)
        else:
            self._append_to_undo_history([value])

    def _load_history(self, before, after, undoing=True):
        """Load a history item and apply it to the array.

        Parameters
        ----------
        before : list of history items
            The list of elements from which we want to load.
        after : list of history items
            The list of element to which to append the loaded element. In the
            case of an undo operation, this is the redo queue, and vice versa.
        undoing : bool
            Whether we are undoing (default) or redoing. In the case of
            redoing, we apply the "after change" element of a history element
            (the third element of the history "atom").

        See Also
        --------
        Labels._save_history
        """
        if len(before) == 0:
            return

        history_item = before.pop()
        after.append(list(reversed(history_item)))
        for prev_indices, prev_values, next_values in reversed(history_item):
            values = prev_values if undoing else next_values
            self.data[prev_indices] = values

        self.refresh()

    def undo(self):
        self._load_history(
            self._undo_history, self._redo_history, undoing=True
        )

    def redo(self):
        self._load_history(
            self._redo_history, self._undo_history, undoing=False
        )

    def fill(self, coord, new_label, refresh=True):
        """Replace an existing label with a new label, either just at the
        connected component if the `contiguous` flag is `True` or everywhere
        if it is `False`, working in the number of dimensions specified by
        the `n_edit_dimensions` flag.

        Parameters
        ----------
        coord : sequence of float
            Position of mouse cursor in image coordinates.
        new_label : int
            Value of the new label to be filled in.
        refresh : bool
            Whether to refresh view slice or not. Set to False to batch paint
            calls.
        """
        int_coord = tuple(np.round(coord).astype(int))
        # If requested fill location is outside data shape then return
        if np.any(np.less(int_coord, 0)) or np.any(
            np.greater_equal(int_coord, self.data.shape)
        ):
            return

        # If requested new label doesn't change old label then return
        old_label = np.asarray(self.data[int_coord]).item()
        if old_label == new_label:
            return
        # If preserve_labels is True, then we only want to fill:
        # - pixels of the background label, filling with the new label
        # - pixels of the previous label, filling with the background
        # the previous label is stored when the selected label is set
        # to the background, e.g. by swap_selected_and_background_labels
        if (
            self.preserve_labels
            and old_label != self._prev_selected_label
            and old_label != self.colormap.background_value
        ):
            return

        dims_to_fill = sorted(
            self._slice_input.order[-self.n_edit_dimensions :]
        )
        data_slice_list = list(int_coord)
        for dim in dims_to_fill:
            data_slice_list[dim] = slice(None)
        data_slice = tuple(data_slice_list)
        labels = np.asarray(self.data[data_slice])
        slice_coord = tuple(int_coord[d] for d in dims_to_fill)

        matches = labels == old_label
        if self.contiguous:
            # if contiguous replace only selected connected component
            labeled_matches, num_features = ndi.label(matches)
            if num_features != 1:
                match_label = labeled_matches[slice_coord]
                matches = np.logical_and(
                    matches, labeled_matches == match_label
                )

        match_indices_local = np.nonzero(matches)
        if self.ndim not in {2, self.n_edit_dimensions}:
            n_idx = len(match_indices_local[0])
            match_indices = []
            j = 0
            for d in data_slice:
                if isinstance(d, slice):
                    match_indices.append(match_indices_local[j])
                    j += 1
                else:
                    match_indices.append(np.full(n_idx, d, dtype=np.intp))
        else:
            match_indices = match_indices_local

        match_indices = _coerce_indices_for_vectorization(
            self.data, match_indices
        )

        self.data_setitem(match_indices, new_label, refresh)

    def _draw(self, new_label, last_cursor_coord, coordinates):
        """Paint into coordinates, accounting for mode and cursor movement.

        The draw operation depends on the current mode of the layer.

        Parameters
        ----------
        new_label : int
            value of label to paint
        last_cursor_coord : sequence
            last painted cursor coordinates
        coordinates : sequence
            new cursor coordinates
        """
        if coordinates is None:
            return
        interp_coord = interpolate_coordinates(
            last_cursor_coord, coordinates, self.brush_size
        )
        for c in interp_coord:
            if (
                self._slice_input.ndisplay == 3
                and self.data[tuple(np.round(c).astype(int))] == 0
            ):
                continue
            if self._mode in [Mode.PAINT, Mode.ERASE]:
                self.paint(c, new_label, refresh=False)
            elif self._mode == Mode.FILL:
                self.fill(c, new_label, refresh=False)
        self._partial_labels_refresh()

    def paint(self, coord, new_label, refresh=True):
        """Paint over existing labels with a new label, using the selected
        brush shape and size, either only on the visible slice or in all
        n dimensions.

        Parameters
        ----------
        coord : sequence of int
            Position of mouse cursor in image coordinates.
        new_label : int
            Value of the new label to be filled in.
        refresh : bool
            Whether to refresh view slice or not. Set to False to batch paint
            calls.
        """
        shape, dims_to_paint = self._get_shape_and_dims_to_paint()
        paint_scale = np.array(
            [self.scale[i] for i in dims_to_paint], dtype=float
        )

        slice_coord = [int(np.round(c)) for c in coord]
        if self.n_edit_dimensions < self.ndim:
            coord_paint = [coord[i] for i in dims_to_paint]
        else:
            coord_paint = coord

        # Ensure circle doesn't have spurious point
        # on edge by keeping radius as ##.5
        radius = np.floor(self.brush_size / 2) + 0.5
        mask_indices = sphere_indices(radius, tuple(paint_scale))

        mask_indices = mask_indices + np.round(np.array(coord_paint)).astype(
            int
        )

        self._paint_indices(
            mask_indices, new_label, shape, dims_to_paint, slice_coord, refresh
        )

    def paint_polygon(self, points, new_label):
        """Paint a polygon over existing labels with a new label.

        Parameters
        ----------
        points : list of coordinates
            List of coordinates of the vertices of a polygon.
        new_label : int
            Value of the new label to be filled in.
        """
        shape, dims_to_paint = self._get_shape_and_dims_to_paint()

        if len(dims_to_paint) != 2:
            raise NotImplementedError(
                'Polygon painting is implemented only in 2D.'
            )

        points = np.array(points, dtype=int)
        slice_coord = points[0].tolist()
        points2d = points[:, dims_to_paint]

        polygon_mask = polygon2mask(shape, points2d)
        mask_indices = np.argwhere(polygon_mask)
        self._paint_indices(
            mask_indices,
            new_label,
            shape,
            dims_to_paint,
            slice_coord,
            refresh=True,
        )

    def _paint_indices(
        self,
        mask_indices,
        new_label,
        shape,
        dims_to_paint,
        slice_coord=None,
        refresh=True,
    ):
        """Paint over existing labels with a new label, using the selected
        mask indices, either only on the visible slice or in all n dimensions.

        Parameters
        ----------
        mask_indices : numpy array of integer coordinates
            Mask to paint represented by an array of its coordinates.
        new_label : int
            Value of the new label to be filled in.
        shape : list
            The label data shape upon which painting is performed.
        dims_to_paint : list
            List of dimensions of the label data that are used for painting.
        refresh : bool
            Whether to refresh view slice or not. Set to False to batch paint
            calls.
        """
        dims_not_painted = sorted(
            self._slice_input.order[: -self.n_edit_dimensions]
        )
        # discard candidate coordinates that are out of bounds
        mask_indices = indices_in_shape(mask_indices, shape)

        # Transfer valid coordinates to slice_coord,
        # or expand coordinate if 3rd dim in 2D image
        slice_coord_temp = list(mask_indices.T)
        if self.n_edit_dimensions < self.ndim:
            for j, i in enumerate(dims_to_paint):
                slice_coord[i] = slice_coord_temp[j]
            for i in dims_not_painted:
                slice_coord[i] = slice_coord[i] * np.ones(
                    mask_indices.shape[0], dtype=int
                )
        else:
            slice_coord = slice_coord_temp

        slice_coord = _coerce_indices_for_vectorization(self.data, slice_coord)

        # slice coord is a tuple of coordinate arrays per dimension
        # subset it if we want to only paint into background/only erase
        # current label, accounting for swap_selected_and_background_labels
        if self.preserve_labels:
            if new_label == self.colormap.background_value:
                keep_coords = self.data[slice_coord] == (
                    self._prev_selected_label
                    if self._prev_selected_label
                    else self.selected_label
                )
            else:
                keep_coords = (
                    self.data[slice_coord] == self.colormap.background_value
                )
            slice_coord = tuple(sc[keep_coords] for sc in slice_coord)

        self.data_setitem(slice_coord, new_label, refresh)

    def _get_shape_and_dims_to_paint(self) -> tuple[list, list]:
        dims_to_paint = sorted(self._get_dims_to_paint())
        shape = list(self.data.shape)

        if self.n_edit_dimensions < self.ndim:
            shape = [shape[i] for i in dims_to_paint]

        return shape, dims_to_paint

    def _get_dims_to_paint(self) -> list:
        return list(self._slice_input.order[-self.n_edit_dimensions :])

    def _get_pt_not_disp(self) -> dict[int, int]:
        """
        Get indices of current visible slice.
        """
        slice_input = self._slice.slice_input
        point = np.round(
            self.world_to_data(slice_input.world_slice.point)
        ).astype(int)
        return {dim: point[dim] for dim in slice_input.not_displayed}

    def data_setitem(self, indices, value, refresh=True):
        """Set `indices` in `data` to `value`, while writing to edit history.

        Parameters
        ----------
        indices : tuple of arrays of int
            Indices in data to overwrite. Must be a tuple of arrays of length
            equal to the number of data dimensions. (Fancy indexing in [2]_).
        value : int or array of int
            New label value(s). If more than one value, must match or
            broadcast with the given indices.
        refresh : bool, default True
            whether to refresh the view, by default True

        References
        ----------
        .. [2] https://numpy.org/doc/stable/user/basics.indexing.html
        """
        changed_indices = self.data[indices] != value
        indices = tuple(x[changed_indices] for x in indices)

        if isinstance(value, Sequence):
            value = np.asarray(value, dtype=self._slice.image.raw.dtype)
        else:
            value = self._slice.image.raw.dtype.type(value)

        # Resize value array to remove unchanged elements
        if isinstance(value, np.ndarray):
            value = value[changed_indices]

        if not indices or indices[0].size == 0:
            return

        self._save_history(
            (
                indices,
                np.array(self.data[indices], copy=True),
                value,
            )
        )

        # update the labels image
        self.data[indices] = value

        pt_not_disp = self._get_pt_not_disp()
        displayed_indices = index_in_slice(
            indices, pt_not_disp, self._slice.slice_input.order
        )
        if isinstance(value, np.ndarray):
            visible_values = value[elements_in_slice(indices, pt_not_disp)]
        else:
            visible_values = value

        if not (  # if not a numpy array or numpy-backed xarray
            isinstance(self.data, np.ndarray)
            or isinstance(getattr(self.data, 'data', None), np.ndarray)
        ):
            # In the absence of slicing, the current slice becomes
            # invalidated by data_setitem; only in the special case of a NumPy
            # array, or a NumPy-array-backed Xarray, is the slice a view and
            # therefore updated automatically.
            # For other types, we update it manually here.
            self._slice.image.raw[displayed_indices] = visible_values

        # tensorstore and xarray do not return their indices in
        # np.ndarray format, so they need to be converted explicitly
        if not isinstance(self.data, np.ndarray):
            indices = [np.array(x).flatten() for x in indices]

        updated_slice = tuple(
            [
                slice(min(axis_indices), max(axis_indices) + 1)
                for axis_indices in indices
            ]
        )

        if self.contour > 0:
            # Expand the slice by 1 pixel as the changes can go beyond
            # the original slice because of the morphological dilation
            # (1 pixel because get_contours always applies 1 pixel dilation)
            updated_slice = expand_slice(updated_slice, self.data.shape, 1)
        else:
            # update data view
            self._slice.image.view[displayed_indices] = (
                self.colormap._data_to_texture(visible_values)
            )

        if self._updated_slice is None:
            self._updated_slice = updated_slice
        else:
            self._updated_slice = tuple(
                [
                    slice(min(s1.start, s2.start), max(s1.stop, s2.stop))
                    for s1, s2 in zip(
                        updated_slice, self._updated_slice, strict=False
                    )
                ]
            )

        if refresh is True:
            self._partial_labels_refresh()

    def _calculate_value_from_ray(self, values):
        non_bg = values != self.colormap.background_value
        if not np.any(non_bg):
            return None
        return values[np.argmax(np.ravel(non_bg))]

    def get_status(
        self,
        position: npt.ArrayLike | None = None,
        *,
        view_direction: npt.ArrayLike | None = None,
        dims_displayed: list[int] | None = None,
        world: bool = False,
        value: Any | None = None,
    ) -> dict[str, str]:
        """Status message information of the data at a coordinate position.

        Parameters
        ----------
        position : tuple
            Position in either data or world coordinates.
        view_direction : Optional[np.ndarray]
            A unit vector giving the direction of the ray in nD world coordinates.
            The default value is None.
        dims_displayed : Optional[List[int]]
            A list of the dimensions currently being displayed in the viewer.
            The default value is None.
        world : bool
            If True the position is taken to be in world coordinates
            and converted into data coordinates. False by default.

        Returns
        -------
        status : dict
            Dict containing a information that can be used in a status update.
        """
        status = super().get_status(
            position,
            view_direction=view_direction,
            dims_displayed=dims_displayed,
            world=world,
        )

        # if this labels layer has properties
        properties = self._get_properties(
            position,
            view_direction=np.asarray(view_direction),
            dims_displayed=dims_displayed,
            world=world,
        )
        if properties:
            status['coordinates'] += '; ' + ', '.join(properties)
            status['value'] += '; ' + ', '.join(properties)

        return status

    def _get_tooltip_text(
        self,
        position,
        *,
        view_direction: np.ndarray | None = None,
        dims_displayed: list[int] | None = None,
        world: bool = False,
    ):
        """
        tooltip message of the data at a coordinate position.

        Parameters
        ----------
        position : tuple
            Position in either data or world coordinates.
        view_direction : Optional[np.ndarray]
            A unit vector giving the direction of the ray in nD world coordinates.
            The default value is None.
        dims_displayed : Optional[List[int]]
            A list of the dimensions currently being displayed in the viewer.
            The default value is None.
        world : bool
            If True the position is taken to be in world coordinates
            and converted into data coordinates. False by default.

        Returns
        -------
        msg : string
            String containing a message that can be used as a tooltip.
        """
        return '\n'.join(
            self._get_properties(
                position,
                view_direction=view_direction,
                dims_displayed=dims_displayed,
                world=world,
            )
        )

    def _get_properties(
        self,
        position,
        *,
        view_direction: np.ndarray | None = None,
        dims_displayed: list[int] | None = None,
        world: bool = False,
    ) -> list:
        if len(self._label_index) == 0 or self.features.shape[1] == 0:
            return []

        value = self.get_value(
            position,
            view_direction=view_direction,
            dims_displayed=dims_displayed,
            world=world,
        )
        # if the cursor is not outside the image or on the background
        if value is None:
            return []

        label_value: int = typing.cast(
            int, value[1] if self.multiscale else value
        )
        if label_value not in self._label_index:
            return [trans._('[No Properties]')]

        idx = self._label_index[label_value]
        return [
            f'{k}: {v[idx]}'
            for k, v in self.features.items()
            if k != 'index'
            and len(v) > idx
            and v[idx] is not None
            and not (isinstance(v[idx], float) and np.isnan(v[idx]))
        ]


def _coerce_indices_for_vectorization(array, indices: list) -> tuple:
    """Coerces indices so that they can be used for vectorized indexing in the given data array."""
    if _is_array_type(array, 'xarray.DataArray'):
        # Fix indexing for xarray if necessary
        # See http://xarray.pydata.org/en/stable/indexing.html#vectorized-indexing
        # for difference from indexing numpy
        try:
            import xarray as xr
        except ModuleNotFoundError:
            pass
        else:
            return tuple(xr.DataArray(i) for i in indices)
    return tuple(indices)



from __future__ import annotations

from collections.abc import Generator, Iterable, Iterator, MutableSet, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

from napari.utils.events import EmitterGroup
from napari.utils.translations import trans

_T = TypeVar('_T')

if TYPE_CHECKING:
    from napari._pydantic_compat import ModelField


class EventedSet(MutableSet[_T]):
    """An unordered collection of unique elements.

    Parameters
    ----------
    data : iterable, optional
        Elements to initialize the set with.

    Events
    ------
    changed (added: Set[_T], removed: Set[_T])
        Emitted when the set changes, includes item(s) that have been added
        and/or removed from the set.
    """

    events: EmitterGroup

    def __init__(self, data: Iterable[_T] = ()) -> None:
        changed = None
        # For inheritance: If the mro already provides an EmitterGroup, add...
        if hasattr(self, 'events') and isinstance(self.events, EmitterGroup):
            self.events.add(changed=changed)
        else:
            # otherwise create a new one
            self.events = EmitterGroup(source=self, changed=changed)

        self._set: set[_T] = set()
        self.update(data)

    # #### START Required Abstract Methods

    def __contains__(self, x: Any) -> bool:
        return x in self._set

    def __iter__(self) -> Iterator[_T]:
        return iter(self._set)

    def __len__(self) -> int:
        return len(self._set)

    def _pre_add_hook(self, value: _T) -> _T:
        # for subclasses to potentially check value before adding
        return value

    def _emit_change(
        self,
        added: set[_T] | None = None,
        removed: set[_T] | None = None,
    ) -> None:
        # provides a hook for subclasses to update internal state before emit
        if added is None:
            added = set()
        if removed is None:
            removed = set()
        self.events.changed(added=added, removed=removed)

    def add(self, value: _T) -> None:
        """Add an element to the set, if not already present."""
        if value not in self:
            value = self._pre_add_hook(value)
            self._set.add(value)
            self._emit_change(added={value}, removed=set())

    def discard(self, value: _T) -> None:
        """Remove an element from a set if it is a member.

        If the element is not a member, do nothing.
        """
        if value in self:
            self._set.discard(value)
            self._emit_change(added=set(), removed={value})

    # #### END Required Abstract Methods

    # methods inherited from Set:
    # __le__, __lt__, __eq__, __ne__, __gt__, __ge__, __and__, __or__,
    # __sub__, __xor__, and isdisjoint

    # methods inherited from MutableSet:
    # clear, pop, remove, __ior__, __iand__, __ixor__, and __isub__

    # The rest are for parity with builtins.set:

    def clear(self) -> None:
        if self._set:
            values = set(self)
            self._set.clear()
            self._emit_change(added=set(), removed=values)

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._set!r})'

    def update(self, others: Iterable[_T] = ()) -> None:
        """Update this set with the union of this set and others"""
        to_add = set(others).difference(self._set)
        if to_add:
            to_add = {self._pre_add_hook(i) for i in to_add}
            self._set.update(to_add)
            self._emit_change(added=set(to_add), removed=set())

    def copy(self) -> EventedSet[_T]:
        """Return a shallow copy of this set."""
        return type(self)(self._set)

    def difference(self, others: Iterable[_T] = ()) -> EventedSet[_T]:
        """Return set of all elements that are in this set but not other."""
        return type(self)(self._set.difference(others))

    def difference_update(self, others: Iterable[_T] = ()) -> None:
        """Remove all elements of another set from this set."""
        to_remove = self._set.intersection(others)
        if to_remove:
            self._set.difference_update(to_remove)
            self._emit_change(added=set(), removed=set(to_remove))

    def intersection(self, others: Iterable[_T] = ()) -> EventedSet[_T]:
        """Return all elements that are in both sets as a new set."""
        return type(self)(self._set.intersection(others))

    def intersection_update(self, others: Iterable[_T] = ()) -> None:
        """Remove all elements of in this set that are not present in other."""
        self.difference_update(self._set.symmetric_difference(others))

    def issubset(self, others: Iterable[_T]) -> bool:
        """Returns whether another set contains this set or not"""
        return self._set.issubset(others)

    def issuperset(self, others: Iterable[_T]) -> bool:
        """Returns whether this set contains another set or not"""
        return self._set.issuperset(others)

    def symmetric_difference(self, others: Iterable[_T]) -> EventedSet[_T]:
        """Returns set of elements that are in exactly one of the sets"""
        return type(self)(self._set.symmetric_difference(others))

    def symmetric_difference_update(self, others: Iterable[_T]) -> None:
        """Update set to the symmetric difference of itself and another.

        This will remove any items in this set that are also in `other`, and
        add any items in others that are not present in this set.
        """
        to_add = set(others).difference(self._set)
        to_remove = self._set.intersection(others)
        self._set.difference_update(to_remove)
        self._set.update(to_add)
        self._emit_change(added=to_add, removed=to_remove)

    def union(self, others: Iterable[_T] = ()) -> EventedSet[_T]:
        """Return a set containing the union of sets"""
        return type(self)(self._set.union(others))

    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.validate

    @classmethod
    def validate(cls, v: Sequence, field: ModelField) -> EventedSet:
        """Pydantic validator."""
        from napari._pydantic_compat import sequence_like

        if not sequence_like(v):
            raise TypeError(
                trans._(
                    'Value is not a valid sequence: {value}',
                    deferred=True,
                    value=v,
                )
            )
        if not field.sub_fields:
            return cls(v)

        type_field = field.sub_fields[0]
        errors = []
        for i, v_ in enumerate(v):
            _valid_value, error = type_field.validate(v_, {}, loc=f'[{i}]')
            if error:
                errors.append(error)
        if errors:
            from napari._pydantic_compat import ValidationError

            raise ValidationError(errors, cls)  # type: ignore [arg-type]
            # need to be fixed when migrate to pydantic 2
        return cls(v)

    def _json_encode(self) -> list:
        """Return an object that can be used by json.dumps."""
        return list(self)



"""PerfTimers class and global instance."""

import contextlib
import os
from collections.abc import Generator
from time import perf_counter_ns

from napari.utils.perf._event import PerfEvent
from napari.utils.perf._stat import Stat
from napari.utils.perf._trace_file import PerfTraceFile

USE_PERFMON = os.getenv('NAPARI_PERFMON', '0') != '0'


class PerfTimers:
    """Timers for performance monitoring.

    Timers are best added using the perfmon config file, which will
    monkey-patch the timers into the code at startup. See
    napari.utils.perf._config for details.

    The collecting timing information can be used in two ways:
    1) Writing a JSON trace file in Chrome's Tracing format.
    2) Napari's real-time QtPerformance widget.

    Attributes
    ----------
    timers : Dict[str, Stat]
        Statistics are kept on each timer.
    trace_file : Optional[PerfTraceFile]
        The tracing file we are writing to if any.

    Notes
    -----
    Chrome deduces nesting based on the start and end times of each timer. The
    chrome://tracing GUI shows the nesting as stacks of colored rectangles.

    However our self.timers dictionary and thus our QtPerformance widget do not
    currently understand nesting. So if they say two timers each took 1ms, you
    can't tell if one called the other or not.

    Despite this limitation when the QtPerformance widget reports slow timers it
    still gives you a good idea what was slow. And then you can use the
    chrome://tracing GUI to see the full story.
    """

    def __init__(self) -> None:
        """Create PerfTimers."""
        # Maps a timer name to one Stat object.
        self.timers: dict[str, Stat] = {}

        # Menu item "Debug -> Record Trace File..." starts a trace.
        self.trace_file: PerfTraceFile | None = None

    def add_event(self, event: PerfEvent) -> None:
        """Save an event to performance trace file and
        update the timers if the event has phase 'X'.

        Parameters
        ----------
        event : PerfEvent
            Add this event.
        """
        # Add event if tracing.
        if self.trace_file is not None:
            self.trace_file.add_event(event)

        if event.phase == 'X':  # Complete Event
            # Update our self.timers (in milliseconds).
            name = event.name
            duration_ms = int(event.duration_ms)
            if name in self.timers:
                self.timers[name].add(duration_ms)
            else:
                self.timers[name] = Stat(duration_ms)

    def add_instant_event(
        self,
        name: str,
        *,
        category: str | None = None,
        process_id: int | None = None,
        thread_id: int | None = None,
        **kwargs: float,
    ) -> None:
        """Build and add one event of length 0.

        Parameters
        ----------
        name : str
            Add this event.
        category : str | None
            Comma separated categories such as "render,update".
        process_id : int | None
            The process id that produced the event.
        thread_id : int | None
            The thread id that produced the event.
        **kwargs
            Arguments to display in the Args section of the Tracing GUI.
        """
        now = perf_counter_ns()
        self.add_event(
            PerfEvent(
                name,
                now,
                now,
                phase='I',
                category=category,
                process_id=process_id,
                thread_id=thread_id,
                **kwargs,
            )
        )

    def add_counter_event(self, name: str, **kwargs: float) -> None:
        """Add one counter event.

        Parameters
        ----------
        name : str
            The name of this event like "draw".
        **kwargs : float
            The individual counters for this event.

        Notes
        -----
        For example add_counter_event("draw", triangles=5, squares=10).
        """
        now = perf_counter_ns()
        self.add_event(
            PerfEvent(
                name,
                now,
                now,
                phase='C',
                category=None,
                process_id=None,
                thread_id=None,
                **kwargs,
            )
        )

    def clear(self) -> None:
        """Clear all timers."""
        # After the GUI displays timing information it clears the timers
        # so that we start accumulating fresh information.
        self.timers.clear()

    def start_trace_file(self, path: str) -> None:
        """Start recording a trace file to disk.

        Parameters
        ----------
        path : str
            Write the trace to this path.
        """
        self.trace_file = PerfTraceFile(path)

    def stop_trace_file(self) -> None:
        """Stop recording a trace file."""
        if self.trace_file is not None:
            self.trace_file.close()
            self.trace_file = None


@contextlib.contextmanager
def block_timer(
    name: str,
    category: str | None = None,
    print_time: bool = False,
    *,
    process_id: int | None = None,
    thread_id: int | None = None,
    phase: str = 'X',
    **kwargs: float,
) -> Generator[PerfEvent, None, None]:
    """Time a block of code.

    block_timer can be used when perfmon is disabled. Use perf_timer instead
    if you want your timer to do nothing when perfmon is disabled.

    Notes
    -----
    Most of the time you should use the perfmon config file to monkey-patch
    perf_timer's into methods and functions. Then you do not need to use
    block_timer or perf_timer context objects explicitly at all.

    Parameters
    ----------
    name : str
        The name of this timer.
    category : str
        Comma separated categories such as "render,update".
    print_time : bool
        Print the duration of the timer when it finishes.
    **kwargs : dict
        Additional keyword arguments for the "args" field of the event.

    Examples
    --------

    .. code-block:: python

        with block_timer("draw") as event:
            draw_stuff()
        print(f"The timer took {event.duration_ms} milliseconds.")

    """
    start_ns = perf_counter_ns()

    # Pass in start_ns for start and end, we call update_end_ns
    # once the block as finished.
    event = PerfEvent(
        name,
        start_ns,
        start_ns,
        category,
        process_id=process_id,
        thread_id=thread_id,
        phase=phase,
        **kwargs,
    )
    yield event

    # Update with the real end time.
    event.update_end_ns(perf_counter_ns())

    if timers:
        timers.add_event(event)
    if print_time:
        print(f'{name} {event.duration_ms:.3f}ms')  # noqa: T201


class DummyTimer:
    """
    An empty timer to use when perfmon is disabled.
    It implements the same interface as PerfTimers.
    """

    def __init__(self) -> None:
        self.trace_file = None

    def add_instant_event(
        self, name: str, **kwargs: str | float | None
    ) -> None:
        """empty timer to use when perfmon is disabled"""

    def add_counter_event(self, name: str, **kwargs: float) -> None:
        """empty timer to use when perfmon is disabled"""

    def add_event(self, event: PerfEvent) -> None:
        """empty timer to use when perfmon is disabled"""

    def start_trace_file(self, path: str) -> None:
        """empty timer to use when perfmon is disabled"""

    def stop_trace_file(self) -> None:
        """empty timer to use when perfmon is disabled"""


def add_instant_event(
    name: str,
    *,
    category: str | None = None,
    process_id: int | None = None,
    thread_id: int | None = None,
    **kwargs: float,
) -> None:
    """Add one instant event.

    Parameters
    ----------
    name : str
        Add this event.
    category : str | None
        Comma separated categories such as "render,update".
    process_id : int | None
        The process id that produced the event.
    thread_id : int | None
        The thread id that produced the event.
    **kwargs
        Arguments to display in the Args section of the Chrome Tracing GUI.
    """
    timers.add_instant_event(
        name,
        category=category,
        process_id=process_id,
        thread_id=thread_id,
        **kwargs,
    )


def add_counter_event(name: str, **kwargs: float) -> None:
    """Add one counter event.

    Parameters
    ----------
    name : str
        The name of this event like "draw".
    **kwargs : float
        The individual counters for this event.

    Notes
    -----
    For example add_counter_event("draw", triangles=5, squares=10).
    """
    timers.add_counter_event(name, **kwargs)


timers: DummyTimer | PerfTimers

if USE_PERFMON:
    timers = PerfTimers()
    perf_timer = block_timer

else:
    # Make sure no one accesses the timers when they are disabled.
    timers = DummyTimer()

    # perf_timer is disabled. Using contextlib.nullcontext did not work.
    @contextlib.contextmanager
    def perf_timer(  # type: ignore [misc]
        name: str,
        category: str | None = None,
        print_time: bool = False,
        *,
        process_id: int | None = None,
        thread_id: int | None = None,
        phase: str = 'X',
        **kwargs: float,
    ) -> Generator[None, None, None]:
        """Do nothing when perfmon is disabled."""
        yield



import numpy as np

from napari._qt.layer_controls.qt_vectors_controls import QtVectorsControls
from napari.layers import Vectors

_VECTORS = np.zeros((2, 2, 2))


def test_out_of_slice_display_checkbox(qtbot):
    layer = Vectors(_VECTORS)
    qtctrl = QtVectorsControls(layer)
    qtbot.addWidget(qtctrl)
    qtctrl.outOfSliceCheckBox.setChecked(True)
    assert layer.out_of_slice_display
    qtctrl.outOfSliceCheckBox.setChecked(False)
    assert not layer.out_of_slice_display



from napari._pydantic_compat import Field
from napari.components.overlays.base import SceneOverlay
from napari.utils.color import ColorValue


class BoundingBoxOverlay(SceneOverlay):
    """
    Bounding box overlay to indicate layer boundaries.

    Attributes
    ----------
    lines : bool
        Whether to show the lines of the bounding box.
    line_thickness : float
        Thickness of the lines in canvas pixels.
    line_color : ColorValue
        Color of the lines.
    points : bool
        Whether to show the vertices of the bounding box as points.
    point_size : float
        Size of the points in canvas pixels.
    point_color : ColorValue
        Color of the points.
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    lines: bool = True
    line_thickness: float = 1
    line_color: ColorValue = Field(default_factory=lambda: ColorValue('red'))
    points: bool = True
    point_size: float = 5
    point_color: ColorValue = Field(default_factory=lambda: ColorValue('blue'))



import os
import sys
from unittest.mock import patch

import pytest
from qtpy.QtCore import QPointF
from qtpy.QtGui import QEnterEvent
from qtpy.QtWidgets import QToolTip

from napari._qt.widgets.qt_tooltip import QtToolTipLabel


@pytest.mark.skipif(
    os.environ.get('CI', False) and sys.platform == 'darwin',
    reason='Timeouts when running on macOS CI',
)
@patch.object(QToolTip, 'showText')
def test_qt_tooltip_label(show_text, qtbot):
    tooltip_text = 'Test QtToolTipLabel showing a tooltip'
    widget = QtToolTipLabel('Label with a tooltip')
    widget.setToolTip(tooltip_text)
    qtbot.addWidget(widget)
    widget.show()

    assert QToolTip.text() == ''
    # simulate movement mouse from outside the widget to the center
    pos = QPointF(widget.rect().center())
    event = QEnterEvent(pos, pos, QPointF(widget.pos()) + pos)
    widget.enterEvent(event)
    assert show_text.called
    assert show_text.call_args[0][1] == tooltip_text



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from app_model.types import KeyBinding
from yaml import SafeDumper, dump_all

from napari._pydantic_compat import BaseModel
from napari.settings._fields import Version

if TYPE_CHECKING:
    from collections.abc import Mapping, Set as AbstractSet
    from typing import Any, TypeVar, Union

    IntStr = Union[int, str]
    AbstractSetIntStr = AbstractSet[IntStr]
    DictStrAny = dict[str, Any]
    MappingIntStrAny = Mapping[IntStr, Any]
    Model = TypeVar('Model', bound=BaseModel)


class YamlDumper(SafeDumper):
    """The default YAML serializer for our pydantic models.

    Add support for custom types by using `YamlDumper.add_representer` or
    `YamlDumper.add_multi_representer` below.
    """


# add_representer requires a strict type match
# add_multi_representer also works for all subclasses of the provided type.
YamlDumper.add_multi_representer(str, YamlDumper.represent_str)
YamlDumper.add_multi_representer(
    Enum, lambda dumper, data: dumper.represent_str(data.value)
)
# the default set representer is ugly:
# disabled_plugins: !!set
#   bioformats: null
# and pydantic will make sure that incoming sets are converted to sets
YamlDumper.add_representer(
    set, lambda dumper, data: dumper.represent_list(data)
)
YamlDumper.add_representer(
    Version, lambda dumper, data: dumper.represent_str(str(data))
)
YamlDumper.add_representer(
    KeyBinding, lambda dumper, data: dumper.represent_str(str(data))
)


class PydanticYamlMixin(BaseModel):
    """Mixin that provides yaml dumping capability to pydantic BaseModel.

    To provide a custom yaml Dumper on a subclass, provide a `yaml_dumper`
    on the Config:

        class Config:
            yaml_dumper = MyDumper
    """

    def yaml(
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny = None,  # type: ignore
        exclude: AbstractSetIntStr | MappingIntStrAny = None,  # type: ignore
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        dumper: type[SafeDumper] | None = None,
        **dumps_kwargs: Any,
    ) -> str:
        """Serialize model to yaml."""
        data = self.dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        if self.__custom_root_type__:
            from napari._pydantic_compat import ROOT_KEY

            data = data[ROOT_KEY]
        return self._yaml_dump(data, dumper, **dumps_kwargs)

    def _yaml_dump(
        self, data, dumper: type[SafeDumper] | None = None, **kw
    ) -> str:
        kw.setdefault('sort_keys', False)
        dumper = dumper or getattr(self.__config__, 'yaml_dumper', YamlDumper)
        return dump_all([data], Dumper=dumper, **kw)



from __future__ import annotations

import logging
import sys
from collections import deque
from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING

from psygnal import Signal

_LOG_SEPARATOR = '<NAPARI_LOG_SEPARATOR>'


if TYPE_CHECKING:
    from typing import Any


if sys.version_info < (3, 11):
    # in python 3.10 there's no public mapping, and this getLevelName function
    # is a bit less resilient (e.g: it "works" both ways (name <-> value), and if
    # invalid values are passed, it just makes up a new level)
    def get_log_level_value(log_level_name: str | None) -> int:
        if log_level_name is None:
            return logging.NOTSET
        return logging.getLevelName(log_level_name)
else:

    def get_log_level_value(log_level_name: str | None) -> int:
        if log_level_name is None:
            return logging.NOTSET
        return logging.getLevelNamesMapping().get(
            log_level_name, logging.NOTSET
        )


class _LogStream:
    """
    Custom stream object to receive logging info.

    Needs to define `write` and `flush` methods which are called by the log handler.
    """

    changed = Signal()

    def __init__(self) -> None:
        self.logs: deque[tuple[Any, ...]] = deque(maxlen=100_000)

    def write(self, log_msg: str) -> None:
        logger, level_name, time, thread, msg = log_msg.split(_LOG_SEPARATOR)
        level_value = get_log_level_value(level_name)
        self.logs.append((logger, level_value, level_name, time, thread, msg))
        # TODO: actually save log to a file somewhere so it can be retrieved?
        self.changed()

    def flush(self) -> None:
        pass

    def get_filtered_logs_html(
        self,
        level: int = logging.DEBUG,
        text_filter: str = '',
        last_only: bool = False,
    ) -> list[str]:
        if isinstance(level, str):
            level = get_log_level_value(level)

        logs = [LOG_STREAM.logs[-1]] if last_only else LOG_STREAM.logs

        selected = [
            (logger_name, level_value, *others)
            for logger_name, level_value, *others in logs
            if level_value >= level
        ]

        # TODO: fuzzy search?
        text_filter = text_filter.lower()
        filtered = [
            log
            for log in selected
            if any(text_filter in str(field).lower() for field in log)
        ]

        return [
            f'{_html_tag_for_level(level_name, level_value)} '
            f'<b>{name}</b> '
            f'<font style="color:gray"><i>[{time}] ({thread})</i></font>: '
            f'{msg}'
            for name, level_value, level_name, time, thread, msg in filtered
        ]


LOG_STREAM = _LogStream()
LOG_HANDLER = logging.StreamHandler(LOG_STREAM)
LOG_HANDLER.setFormatter(
    logging.Formatter(
        f'%(name)s{_LOG_SEPARATOR}%(levelname)s{_LOG_SEPARATOR}%(asctime)s{_LOG_SEPARATOR}%(threadName)s{_LOG_SEPARATOR}%(message)s'
    )
)
LOG_HANDLER.setLevel(logging.DEBUG)


@contextmanager
def register_logger_to_napari_handler(
    module: str,
) -> Generator[None, None, None]:
    """
    Register a specific module's logger to use our custom log handler.
    """
    logger = logging.getLogger(module)
    # ensure the default "last resort" logging to console remains
    if not logger.handlers and logging.lastResort:
        logger.addHandler(logging.lastResort)
    logger.addHandler(LOG_HANDLER)
    yield
    logger.removeHandler(LOG_HANDLER)


def _html_tag_for_level(level_name: str, level_value: int) -> str:
    """
    Generate html tag for the appropriate logging level.
    """
    colors = {
        logging.INFO: 'cyan',
        logging.WARNING: 'orange',
        logging.ERROR: 'red',
        logging.CRITICAL: 'magenta',
    }
    color = 'blue'
    for level, level_color in colors.items():
        if level_value >= level:
            color = level_color
    # this is ugly AF but html is weird and I don't get it
    padding = '&nbsp;' * (8 - len(level_name))
    return f'<font style="color:{color}">{level_name}{padding}</font>'



import os
from collections import defaultdict
from unittest.mock import Mock

import pytest
from qtpy.QtWidgets import QAction, QShortcut

from napari._qt.qt_event_loop import (
    _ipython_has_eventloop,
    run,
    set_app_id,
)


@pytest.mark.skipif(os.name != 'nt', reason='Windows specific')
def test_windows_grouping_overwrite(qapp):
    import ctypes

    def get_app_id():
        mem = ctypes.POINTER(ctypes.c_wchar)()
        ctypes.windll.shell32.GetCurrentProcessExplicitAppUserModelID(
            ctypes.byref(mem)
        )
        res = ctypes.wstring_at(mem)
        ctypes.windll.Ole32.CoTaskMemFree(mem)
        return res

    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID('test_text')

    assert get_app_id() == 'test_text'
    set_app_id('custom_string')
    assert get_app_id() == 'custom_string'
    set_app_id('')  # app id can't be an empty string
    assert get_app_id() == 'custom_string'
    set_app_id(' ')
    assert get_app_id() == ' '


def test_run_outside_ipython(make_napari_viewer, qapp, monkeypatch):
    """Test that we don't incorrectly give ipython the event loop."""
    assert not _ipython_has_eventloop()
    v1 = make_napari_viewer()
    assert not _ipython_has_eventloop()
    v2 = make_napari_viewer()
    assert not _ipython_has_eventloop()

    with monkeypatch.context() as m:
        mock_exec = Mock()
        m.setattr(qapp, 'exec_', mock_exec)
        run()
        mock_exec.assert_called_once()

    v1.close()
    v2.close()


def test_shortcut_collision(qtbot, make_napari_viewer):
    viewer = make_napari_viewer()
    defined_shortcuts = defaultdict(list)
    problematic_shortcuts = []
    shortcuts = viewer.window._qt_window.findChildren(QShortcut)
    for shortcut in shortcuts:
        key = shortcut.key().toString()
        if key == 'Ctrl+M':
            # menubar toggle support
            # https://github.com/napari/napari/pull/3204
            continue
        if key and key in defined_shortcuts:
            problematic_shortcuts.append(key)
        defined_shortcuts[key].append(key)

    actions = viewer.window._qt_window.findChildren(QAction)
    for action in actions:
        key = action.shortcut().toString()
        if key and key in defined_shortcuts:
            problematic_shortcuts.append(key)
        defined_shortcuts[key].append(key)
    assert not problematic_shortcuts
    # due to throttled mouse_move, a timer is started by the viewer, so we
    # need to wait for it to be done
    qtbot.wait(10)



from collections import defaultdict
from collections.abc import MutableMapping, Sequence
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    cast,
    overload,
)
from warnings import warn

import numpy as np
from typing_extensions import Self

from napari._pydantic_compat import Field, PrivateAttr, validator
from napari.utils.color import ColorArray
from napari.utils.colormaps import _accelerated_cmap as _accel_cmap
from napari.utils.colormaps.colorbars import make_colorbar
from napari.utils.colormaps.standardize_color import transform_color
from napari.utils.compat import StrEnum
from napari.utils.events import EventedModel
from napari.utils.events.custom_types import Array
from napari.utils.migrations import deprecated_class_name
from napari.utils.translations import trans

if TYPE_CHECKING:
    from numba import typed


class ColormapInterpolationMode(StrEnum):
    """INTERPOLATION: Interpolation mode for colormaps.

    Selects an interpolation mode for the colormap.
            * linear: colors are defined by linear interpolation between
              colors of neighboring controls points.
            * zero: colors are defined by the value of the color in the
              bin between by neighboring controls points.
    """

    LINEAR = 'linear'
    ZERO = 'zero'


class Colormap(EventedModel):
    """Colormap that relates intensity values to colors.

    Attributes
    ----------
    colors : array, shape (N, 4)
        Data used in the colormap.
    name : str
        Name of the colormap.
    _display_name : str
        Display name of the colormap.
    controls : array, shape (N,) or (N+1,)
        Control points of the colormap.
    interpolation : str
        Colormap interpolation mode, either 'linear' or
        'zero'. If 'linear', ncontrols = ncolors (one
        color per control point). If 'zero', ncontrols
        = ncolors+1 (one color per bin).
    """

    # fields
    colors: ColorArray
    name: str = 'custom'
    _display_name: str | None = PrivateAttr(None)
    interpolation: ColormapInterpolationMode = ColormapInterpolationMode.LINEAR
    controls: Array = Field(default_factory=lambda: cast(Array, []))

    def __init__(
        self, colors, display_name: str | None = None, **data
    ) -> None:
        if display_name is None:
            display_name = data.get('name', 'custom')

        super().__init__(colors=colors, **data)
        self._display_name = display_name

    # controls validator must be called even if None for correct initialization
    @validator('controls', pre=True, always=True, allow_reuse=True)
    def _check_controls(cls, v, values):
        # If no control points provided generate defaults
        if v is None or len(v) == 0:
            n_controls = len(values['colors']) + int(
                values['interpolation'] == ColormapInterpolationMode.ZERO
            )
            return np.linspace(0, 1, n_controls, dtype=np.float32)

        # Check control end points are correct
        if v[0] != 0 or (len(v) > 1 and v[-1] != 1):
            raise ValueError(
                trans._(
                    'Control points must start with 0.0 and end with 1.0. '
                    'Got {start_control_point} and {end_control_point}',
                    deferred=True,
                    start_control_point=v[0],
                    end_control_point=v[-1],
                )
            )

        # Check control points are sorted correctly
        if not np.array_equal(v, sorted(v)):
            raise ValueError(
                trans._(
                    'Control points need to be sorted in ascending order',
                    deferred=True,
                )
            )

        # Check number of control points is correct
        n_controls_target = len(values.get('colors', [])) + int(
            values['interpolation'] == ColormapInterpolationMode.ZERO
        )
        n_controls = len(v)
        if n_controls != n_controls_target:
            raise ValueError(
                trans._(
                    'Wrong number of control points provided. Expected {n_controls_target}, got {n_controls}',
                    deferred=True,
                    n_controls_target=n_controls_target,
                    n_controls=n_controls,
                )
            )

        return v

    def __iter__(self):
        yield from (self.colors, self.controls, self.interpolation)

    def __len__(self):
        return len(self.colors)

    def map(self, values):
        values = np.atleast_1d(values)
        if self.interpolation == ColormapInterpolationMode.LINEAR:
            # One color per control point
            cols = [
                np.interp(values, self.controls, self.colors[:, i])
                for i in range(4)
            ]
            cols = np.stack(cols, axis=-1)
        elif self.interpolation == ColormapInterpolationMode.ZERO:
            # One color per bin
            # Colors beyond max clipped to final bin
            indices = np.clip(
                np.searchsorted(self.controls, values, side='right') - 1,
                0,
                len(self.colors) - 1,
            )
            cols = self.colors[indices.astype(np.int32)]
        else:
            raise ValueError(
                trans._(
                    'Unrecognized Colormap Interpolation Mode',
                    deferred=True,
                )
            )

        return cols

    @property
    def colorbar(self):
        return make_colorbar(self)


class LabelColormapBase(Colormap):
    use_selection: bool = False
    selection: int = 0
    background_value: int = 0
    interpolation: Literal[ColormapInterpolationMode.ZERO] = Field(
        ColormapInterpolationMode.ZERO, frozen=True
    )
    _cache_mapping: dict[tuple[np.dtype, np.dtype], np.ndarray] = PrivateAttr(
        default={}
    )
    _cache_other: dict[str, Any] = PrivateAttr(default={})

    class Config(Colormap.Config):
        # this config is to avoid deepcopy of cached_property
        # see https://github.com/pydantic/pydantic/issues/2763
        # it is required until we drop Pydantic 1 or Python 3.11 and older
        # need to validate after drop pydantic 1
        keep_untouched = (cached_property,)

    @overload
    def _data_to_texture(self, values: np.ndarray) -> np.ndarray: ...

    @overload
    def _data_to_texture(self, values: np.integer) -> np.integer: ...

    def _data_to_texture(
        self, values: np.ndarray | np.integer
    ) -> np.ndarray | np.integer:
        """Map input values to values for send to GPU."""
        raise NotImplementedError

    def _cmap_without_selection(self) -> Self:
        if self.use_selection:
            cmap = self.__class__(**self.dict())
            cmap.use_selection = False
            return cmap
        return self

    def _get_mapping_from_cache(
        self, data_dtype: np.dtype
    ) -> np.ndarray | None:
        """For given dtype, return precomputed array mapping values to colors.

        Returns None if the dtype itemsize is greater than 2.
        """
        target_dtype = _texture_dtype(self._num_unique_colors, data_dtype)
        key = (data_dtype, target_dtype)
        if key not in self._cache_mapping and data_dtype.itemsize <= 2:
            data = np.arange(
                np.iinfo(target_dtype).max + 1, dtype=target_dtype
            ).astype(data_dtype)
            self._cache_mapping[key] = self._map_without_cache(data)
        return self._cache_mapping.get(key)

    def _clear_cache(self):
        """Mechanism to clean cached properties"""
        self._cache_mapping = {}
        self._cache_other = {}

    @property
    def _num_unique_colors(self) -> int:
        """Number of unique colors, not counting transparent black."""
        return len(self.colors) - 1

    def _map_without_cache(self, values: np.ndarray) -> np.ndarray:
        """Function that maps values to colors without selection or cache"""
        raise NotImplementedError

    def _selection_as_minimum_dtype(self, dtype: np.dtype) -> int:
        """Treat selection as given dtype and calculate value with min dtype.

        Parameters
        ----------
        dtype : np.dtype
            The dtype to convert the selection to.

        Returns
        -------
        int
            The selection converted.
        """
        return int(self._data_to_texture(dtype.type(self.selection)))


class CyclicLabelColormap(LabelColormapBase):
    """Color cycle with a background value.

    Attributes
    ----------
    colors : ColorArray
        Colors to be used for mapping.
        For values above the number of colors,
        the colors will be cycled.
    use_selection : bool
        Whether map only selected label.
        If `True` only selected label will be mapped to not transparent color.
    selection : int
        The selected label.
    background_value : int
        Which value should be treated as a background
        and mapped to transparent color.
    interpolation : Literal['zero']
        required by implementation, please do not set value
    seed : float
        seed used for random color generation. Used for reproducibility.
        It will be removed in the future release.
    """

    seed: float = 0.5

    @validator('colors', allow_reuse=True)
    def _validate_color(cls, v):
        if len(v) > 2**16:
            raise ValueError(
                'Only up to 2**16=65535 colors are supported for LabelColormap'
            )
        return v

    def _background_as_minimum_dtype(self, dtype: np.dtype) -> int:
        """Treat background as given dtype and calculate value with min dtype.

        Parameters
        ----------
        dtype : np.dtype
            The dtype to convert the background to.

        Returns
        -------
        int
            The background converted.
        """
        return int(self._data_to_texture(dtype.type(self.background_value)))

    @overload
    def _data_to_texture(self, values: np.ndarray) -> np.ndarray: ...

    @overload
    def _data_to_texture(self, values: np.integer) -> np.integer: ...

    def _data_to_texture(
        self, values: np.ndarray | np.integer
    ) -> np.ndarray | np.integer:
        """Map input values to values for send to GPU."""
        return _cast_labels_data_to_texture_dtype_auto(values, self)

    def _map_without_cache(self, values) -> np.ndarray:
        texture_dtype_values = _accel_cmap.zero_preserving_modulo_numpy(
            values,
            len(self.colors) - 1,
            values.dtype,
            self.background_value,
        )
        mapped = self.colors[texture_dtype_values]
        mapped[texture_dtype_values == 0] = 0
        return mapped

    def map(self, values: np.ndarray | np.integer | int) -> np.ndarray:
        """Map values to colors.

        Parameters
        ----------
        values : np.ndarray or int
            Values to be mapped.

        Returns
        -------
        np.ndarray of the same shape as values,
            but with the last dimension of size 4
            Mapped colors.
        """
        original_shape = np.shape(values)
        values = np.atleast_1d(values)

        if values.dtype.kind == 'f':
            values = values.astype(np.int64)
        mapper = self._get_mapping_from_cache(values.dtype)
        if mapper is not None:
            mapped = mapper[values]
        else:
            mapped = self._map_without_cache(values)
        if self.use_selection:
            mapped[(values != self.selection)] = 0

        return np.reshape(mapped, original_shape + (4,))

    def shuffle(self, seed: int):
        """Shuffle the colormap colors.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.
        """
        np.random.default_rng(seed).shuffle(self.colors[1:])
        self.events.colors(value=self.colors)


LabelColormap = deprecated_class_name(
    CyclicLabelColormap,
    'LabelColormap',
    version='0.5.0',
    since_version='0.4.19',
)


class DirectLabelColormap(LabelColormapBase):
    """Colormap using a direct mapping from labels to color using a dict.

    Attributes
    ----------
    color_dict: dict from int to (3,) or (4,) array
        The dictionary mapping labels to colors.

    use_selection : bool
        Whether to map only the selected label to a color.
        If `True` only selected label will be not transparent.
    selection : int
        The selected label.
    colors : ColorArray
        Exist because of implementation details. Please do not use it.
    """

    color_dict: defaultdict[int | None, np.ndarray] = Field(
        default_factory=lambda: defaultdict(lambda: np.zeros(4))
    )
    use_selection: bool = False
    selection: int = 0

    def __init__(self, *args, **kwargs) -> None:
        if 'colors' not in kwargs and not args:
            kwargs['colors'] = np.zeros(3)
        super().__init__(*args, **kwargs)

    def __len__(self):
        """Overwrite from base class because .color is a dummy array.

        This returns the number of colors in the colormap, including
        background and unmapped labels.
        """
        return self._num_unique_colors + 2

    @validator('color_dict', pre=True, always=True, allow_reuse=True)
    def _validate_color_dict(cls, v, values):
        """Ensure colors are RGBA arrays, not strings.

        Parameters
        ----------
        cls : type
            The class of the object being instantiated.
        v : MutableMapping
            A mapping from integers to colors. It *may* have None as a key,
            which indicates the color to map items not in the dictionary.
            Alternatively, it could be a defaultdict. If neither is provided,
            missing colors are not rendered (rendered as fully transparent).
        values : dict[str, Any]
            A dictionary mapping previously-validated attributes to their
            validated values. Attributes are validated in the order in which
            they are defined.

        Returns
        -------
        res : (default)dict[int, np.ndarray[float]]
            A properly-formatted dictionary mapping labels to RGBA arrays.
        """
        if not isinstance(v, defaultdict) and None not in v:
            warn(
                'color_dict did not provide a default color. '
                'Missing keys will be transparent. '
                'To provide a default color, use the key `None`, '
                'or provide a defaultdict instance.'
            )
            v = {**v, None: 'transparent'}
        res = {
            label: transform_color(color_str)[0]
            for label, color_str in v.items()
        }
        if (
            'background_value' in values
            and (bg := values['background_value']) not in res
        ):
            res[bg] = transform_color('transparent')[0]
        if isinstance(v, defaultdict):
            res = defaultdict(v.default_factory, res)
        return res

    def _selection_as_minimum_dtype(self, dtype: np.dtype) -> int:
        return int(
            _cast_labels_data_to_texture_dtype_direct(
                dtype.type(self.selection), self
            )
        )

    @overload
    def _data_to_texture(self, values: np.ndarray) -> np.ndarray: ...

    @overload
    def _data_to_texture(self, values: np.integer) -> np.integer: ...

    def _data_to_texture(
        self, values: np.ndarray | np.integer
    ) -> np.ndarray | np.integer:
        """Map input values to values for send to GPU."""
        return _cast_labels_data_to_texture_dtype_direct(values, self)

    def map(self, values: np.ndarray | np.integer | int) -> np.ndarray:
        """Map values to colors.

        Parameters
        ----------
        values : np.ndarray or int
            Values to be mapped.

        Returns
        -------
        np.ndarray of same shape as values, but with last dimension of size 4
            Mapped colors.
        """
        if isinstance(values, np.integer):
            values = int(values)
        if isinstance(values, int):
            if self.use_selection and values != self.selection:
                return np.array((0, 0, 0, 0))
            return self.color_dict.get(values, self.default_color)
        if isinstance(values, list | tuple):
            values = np.array(values)
        if not isinstance(values, np.ndarray) or values.dtype.kind in 'fU':
            raise TypeError('DirectLabelColormap can only be used with int')
        mapper = self._get_mapping_from_cache(values.dtype)
        if mapper is not None:
            mapped = mapper[values]
        else:
            values_cast = _accel_cmap.labels_raw_to_texture_direct(
                values, self
            )
            mapped = self._map_precast(values_cast, apply_selection=True)

        if self.use_selection:
            mapped[(values != self.selection)] = 0
        return mapped

    def _map_without_cache(self, values: np.ndarray) -> np.ndarray:
        cmap = self._cmap_without_selection()
        cast = _accel_cmap.labels_raw_to_texture_direct(values, cmap)
        return self._map_precast(cast, apply_selection=False)

    def _map_precast(self, values, apply_selection) -> np.ndarray:
        """Map values to colors.

        Parameters
        ----------
        values : np.ndarray
            Values to be mapped. It need to be already cast using
            cast_labels_to_minimum_type_auto

        Returns
        -------
        np.ndarray of shape (N, M, 4)
            Mapped colors.

        Notes
        -----
        it is implemented for thumbnail labels,
        where we already have cast values
        """
        mapped = np.zeros(values.shape + (4,), dtype=np.float32)
        colors = self._values_mapping_to_minimum_values_set(apply_selection)[1]
        for idx in np.ndindex(values.shape):
            value = values[idx]
            mapped[idx] = colors[value]
        return mapped

    @cached_property
    def _num_unique_colors(self) -> int:
        """Count the number of unique colors in the colormap.

        This number does not include background or the default color for
        unmapped labels.
        """
        return len({tuple(x) for x in self.color_dict.values()})

    def _clear_cache(self):
        super()._clear_cache()
        if '_num_unique_colors' in self.__dict__:
            del self.__dict__['_num_unique_colors']
        if '_label_mapping_and_color_dict' in self.__dict__:
            del self.__dict__['_label_mapping_and_color_dict']
        if '_array_map' in self.__dict__:
            del self.__dict__['_array_map']

    def _values_mapping_to_minimum_values_set(
        self, apply_selection=True
    ) -> tuple[dict[int | None, int], dict[int, np.ndarray]]:
        """Create mapping from original values to minimum values set.
        To use minimum possible dtype for labels.

        Returns
        -------
        Dict[Optional[int], int]
            Mapping from original values to minimum values set.
        Dict[int, np.ndarray]
            Mapping from new values to colors.

        """
        if self.use_selection and apply_selection:
            return {self.selection: 1, None: 0}, {
                0: np.array((0, 0, 0, 0)),
                1: self.color_dict.get(
                    self.selection,
                    self.default_color,
                ),
            }

        return self._label_mapping_and_color_dict

    @cached_property
    def _label_mapping_and_color_dict(
        self,
    ) -> tuple[dict[int | None, int], dict[int, np.ndarray]]:
        color_to_labels: dict[tuple[int, ...], list[int | None]] = {}
        labels_to_new_labels: dict[int | None, int] = {
            None: _accel_cmap.MAPPING_OF_UNKNOWN_VALUE
        }
        new_color_dict: dict[int, np.ndarray] = {
            _accel_cmap.MAPPING_OF_UNKNOWN_VALUE: self.default_color,
        }

        for label, color in self.color_dict.items():
            if label is None:
                continue
            color_tup = tuple(color)
            if color_tup not in color_to_labels:
                color_to_labels[color_tup] = [label]
                labels_to_new_labels[label] = len(new_color_dict)
                new_color_dict[labels_to_new_labels[label]] = color
            else:
                color_to_labels[color_tup].append(label)
                labels_to_new_labels[label] = labels_to_new_labels[
                    color_to_labels[color_tup][0]
                ]

        return labels_to_new_labels, new_color_dict

    def _get_typed_dict_mapping(self, data_dtype: np.dtype) -> 'typed.Dict':
        """Create mapping from label values to texture values of smaller dtype.

        In https://github.com/napari/napari/issues/6397, we noticed that using
        float32 textures was much slower than uint8 or uint16 textures. When
        labels data is (u)int(8,16), we simply use the labels data directly.
        But when it is higher-precision, we need to compress the labels into
        the smallest dtype that can still achieve the goal of the
        visualisation. This corresponds to the smallest dtype that can map to
        the number of unique colors in the colormap. Even if we have a
        million labels, if they map to one of two colors, we can map them to
        a uint8 array with values 1 and 2; then, the texture can map those
        two values to each of the two possible colors.

        Returns
        -------
        Dict[Optional[int], int]
            Mapping from original values to minimal texture value set.
        """

        # we cache the result to avoid recomputing it on each slice;
        # check first if it's already in the cache.
        key = f'_{data_dtype}_typed_dict'
        if key in self._cache_other:
            return self._cache_other[key]

        from numba import typed, types

        # num_unique_colors + 2 because we need to map None and background
        target_type = _accel_cmap.minimum_dtype_for_labels(
            self._num_unique_colors + 2
        )

        dkt = typed.Dict.empty(
            key_type=getattr(types, data_dtype.name),
            value_type=getattr(types, target_type.name),
        )
        iinfo = np.iinfo(data_dtype)
        for k, v in self._label_mapping_and_color_dict[0].items():
            # ignore values outside the data dtype, since they will never need
            # to be colormapped from that dtype.
            if k is not None and iinfo.min <= k <= iinfo.max:
                dkt[data_dtype.type(k)] = target_type.type(v)

        self._cache_other[key] = dkt

        return dkt

    @cached_property
    def _array_map(self):
        """Create an array to map labels to texture values of smaller dtype."""

        max_value = max(
            (abs(x) for x in self.color_dict if x is not None), default=0
        )
        if any(x < 0 for x in self.color_dict if x is not None):
            max_value *= 2
        if max_value > 2**16:
            raise RuntimeError(  # pragma: no cover
                'Cannot use numpy implementation for large values of labels '
                'direct colormap. Please install numba.'
            )
        dtype = _accel_cmap.minimum_dtype_for_labels(
            self._num_unique_colors + 2
        )
        label_mapping = self._values_mapping_to_minimum_values_set()[0]

        # We need 2 + the max value: one because we will be indexing with the
        # max value, and an extra one so that higher values get clipped to
        # that index and map to the default value, rather than to the max
        # value in the map.
        mapper = np.full(
            (max_value + 2), _accel_cmap.MAPPING_OF_UNKNOWN_VALUE, dtype=dtype
        )
        for key, val in label_mapping.items():
            if key is None:
                continue
            mapper[key] = val
        return mapper

    @property
    def default_color(self) -> np.ndarray:
        return self.color_dict.get(None, np.array((0, 0, 0, 0)))
        # we provided here default color for backward compatibility
        # if someone is using DirectLabelColormap directly, not through Label layer


@overload
def _convert_small_ints_to_unsigned(
    data: np.ndarray,
) -> np.ndarray: ...


@overload
def _convert_small_ints_to_unsigned(
    data: np.integer,
) -> np.integer: ...


def _convert_small_ints_to_unsigned(
    data: np.ndarray | np.integer,
) -> np.ndarray | np.integer:
    """Convert (u)int8 to uint8 and (u)int16 to uint16.

    Otherwise, return the original array.

    Parameters
    ----------
    data : np.ndarray | np.integer
        Data to be converted.

    Returns
    -------
    np.ndarray | np.integer
        Converted data.
    """
    if data.dtype.itemsize == 1:
        # for fast rendering of int8
        return data.view(np.uint8)
    if data.dtype.itemsize == 2:
        # for fast rendering of int16
        return data.view(np.uint16)
    return data


@overload
def _cast_labels_data_to_texture_dtype_auto(
    data: np.ndarray,
    colormap: CyclicLabelColormap,
) -> np.ndarray: ...


@overload
def _cast_labels_data_to_texture_dtype_auto(
    data: np.integer,
    colormap: CyclicLabelColormap,
) -> np.integer: ...


def _cast_labels_data_to_texture_dtype_auto(
    data: np.ndarray | np.integer,
    colormap: CyclicLabelColormap,
) -> np.ndarray | np.integer:
    """Convert labels data to the data type used in the texture.

    In https://github.com/napari/napari/issues/6397, we noticed that using
    float32 textures was much slower than uint8 or uint16 textures. Here we
    convert the labels data to uint8 or uint16, based on the following rules:

    - uint8 and uint16 labels data are unchanged. (No copy of the arrays.)
    - int8 and int16 data are converted with a *view* to uint8 and uint16.
      (This again does not involve a copy so is fast, and lossless.)
    - higher precision integer data (u)int{32,64} are hashed to uint8, uint16,
      or float32, depending on the number of colors in the input colormap. (See
      `minimum_dtype_for_labels`.) Since the hashing can result in collisions,
      this conversion *has* to happen in the CPU to correctly map the
      background and selection values.

    Parameters
    ----------
    data : np.ndarray
        Labels data to be converted.
    colormap : CyclicLabelColormap
        Colormap used to display the labels data.

    Returns
    -------
    np.ndarray | np.integer
        Converted labels data.
    """
    original_shape = np.shape(data)
    if data.itemsize <= 2:
        return _convert_small_ints_to_unsigned(data)

    data_arr = np.atleast_1d(data)
    num_colors = len(colormap.colors) - 1
    zero_preserving_modulo_func = _accel_cmap.zero_preserving_modulo
    if isinstance(data, np.integer):
        zero_preserving_modulo_func = _accel_cmap.zero_preserving_modulo_numpy

    dtype = _accel_cmap.minimum_dtype_for_labels(num_colors + 1)

    if colormap.use_selection:
        selection_in_texture = _accel_cmap.zero_preserving_modulo_numpy(
            np.array([colormap.selection]), num_colors, dtype
        )
        converted = np.where(
            data_arr == colormap.selection, selection_in_texture, dtype.type(0)
        )
    else:
        converted = zero_preserving_modulo_func(
            data_arr, num_colors, dtype, colormap.background_value
        )

    if isinstance(data, np.integer):
        return dtype.type(converted[0])

    return np.reshape(converted, original_shape)


@overload
def _cast_labels_data_to_texture_dtype_direct(
    data: np.ndarray, direct_colormap: DirectLabelColormap
) -> np.ndarray: ...


@overload
def _cast_labels_data_to_texture_dtype_direct(
    data: np.integer, direct_colormap: DirectLabelColormap
) -> np.integer: ...


def _cast_labels_data_to_texture_dtype_direct(
    data: np.ndarray | np.integer, direct_colormap: DirectLabelColormap
) -> np.ndarray | np.integer:
    """Convert labels data to the data type used in the texture.

    In https://github.com/napari/napari/issues/6397, we noticed that using
    float32 textures was much slower than uint8 or uint16 textures. Here we
    convert the labels data to uint8 or uint16, based on the following rules:

    - uint8 and uint16 labels data are unchanged. (No copy of the arrays.)
    - int8 and int16 data are converted with a *view* to uint8 and uint16.
      (This again does not involve a copy so is fast, and lossless.)
    - higher precision integer data (u)int{32,64} are mapped to an intermediate
      space of sequential values based on the colors they map to. As an
      example, if the values are [1, 2**25, and 2**50], and the direct
      colormap maps them to ['red', 'green', 'red'], then the intermediate map
      is {1: 1, 2**25: 2, 2**50: 1}. The labels can then be converted to a
      uint8 texture and a smaller direct colormap with only two values.

    This function calls `_labels_raw_to_texture_direct`, but makes sure that
    signed ints are first viewed as their unsigned counterparts.

    Parameters
    ----------
    data : np.ndarray | np.integer
        Labels data to be converted.
    direct_colormap : CyclicLabelColormap
        Colormap used to display the labels data.

    Returns
    -------
    np.ndarray | np.integer
        Converted labels data.
    """
    data = _convert_small_ints_to_unsigned(data)

    if data.itemsize <= 2:
        return data

    if isinstance(data, np.integer):
        mapper = direct_colormap._label_mapping_and_color_dict[0]
        target_dtype = _accel_cmap.minimum_dtype_for_labels(
            direct_colormap._num_unique_colors + 2
        )
        return target_dtype.type(
            mapper.get(int(data), _accel_cmap.MAPPING_OF_UNKNOWN_VALUE)
        )

    original_shape = np.shape(data)
    array_data = np.atleast_1d(data)
    return np.reshape(
        _accel_cmap.labels_raw_to_texture_direct(array_data, direct_colormap),
        original_shape,
    )


def _texture_dtype(num_colors: int, dtype: np.dtype) -> np.dtype:
    """Compute VisPy texture dtype given number of colors and raw data dtype.

    - for data of type int8 and uint8 we can use uint8 directly, with no copy.
    - for int16 and uint16 we can use uint16 with no copy.
    - for any other dtype, we fall back on `minimum_dtype_for_labels`, which
      will require on-CPU mapping between the raw data and the texture dtype.
    """
    if dtype.itemsize == 1:
        return np.dtype(np.uint8)
    if dtype.itemsize == 2:
        return np.dtype(np.uint16)
    return _accel_cmap.minimum_dtype_for_labels(num_colors)


def _normalize_label_colormap(
    any_colormap_like,
) -> CyclicLabelColormap | DirectLabelColormap:
    """Convenience function to convert color containers to LabelColormaps.

    A list of colors or 2D nparray of colors is interpreted as a color cycle
    (`CyclicLabelColormap`), and a mapping of colors is interpreted as a direct
    color map (`DirectLabelColormap`).

    Parameters
    ----------
    any_colormap_like : Sequence[color], MutableMapping[int, color], ...
        An object that can be interpreted as a LabelColormap, including a
        LabelColormap directly.

    Returns
    -------
    CyclicLabelColormap | DirectLabelColormap
        The computed LabelColormap object.
    """
    if isinstance(
        any_colormap_like, CyclicLabelColormap | DirectLabelColormap
    ):
        return any_colormap_like
    if isinstance(any_colormap_like, Sequence):
        return CyclicLabelColormap(any_colormap_like)
    if isinstance(any_colormap_like, MutableMapping):
        return DirectLabelColormap(color_dict=any_colormap_like)
    if (
        isinstance(any_colormap_like, np.ndarray)
        and any_colormap_like.ndim == 2
        and any_colormap_like.shape[1] in (3, 4)
    ):
        return CyclicLabelColormap(any_colormap_like)
    raise ValueError(
        f'Unable to interpret as labels colormap: {any_colormap_like}'
    )



from dataclasses import dataclass, field

import pytest

from napari.utils._register import create_func
from napari.utils.migrations import (
    deprecated_constructor_arg_by_attr,
    rename_argument,
)


@dataclass
class DummyClass:
    """Dummy class to test create_func"""

    layers: list = field(default_factory=list)


class SimpleClass:
    """Simple class to test create_func"""

    def __init__(self, a):
        self.a = a


class SimpleClassDeprecated:
    """Simple class to test create_func"""

    @deprecated_constructor_arg_by_attr('b')
    @deprecated_constructor_arg_by_attr('c')
    def __init__(self, a=1):
        self.a = a

    @property
    def b(self):  # pragma: no cover
        return self.a * 2

    @b.setter
    def b(self, value):
        self.a = value // 2

    @property
    def c(self):  # pragma: no cover
        return self.a * 2

    @c.setter
    def c(self, value):
        self.a = value // 2


class SimpleClassRenamed:
    """Simple class to test create_func"""

    @rename_argument(
        from_name='c',
        to_name='a',
        version='0.6.0',
        since_version='0.4.18',
    )
    @rename_argument(
        from_name='d',
        to_name='b',
        version='0.6.0',
        since_version='0.4.18',
    )
    def __init__(self, a=1, b=1):
        self.a = a
        self.b = b


def test_create_func():
    DummyClass.add_simple_class = create_func(SimpleClass)
    dc = DummyClass()
    dc.add_simple_class(a=1)
    assert dc.layers[0].a == 1


def test_create_func_deprecated():
    DummyClass.add_simple_class_deprecated = create_func(SimpleClassDeprecated)
    dc = DummyClass()
    dc.add_simple_class_deprecated(b=4)
    assert dc.layers[0].a == 2
    dc.add_simple_class_deprecated(c=8)
    assert dc.layers[1].a == 4


def test_create_func_renamed():
    DummyClass.add_simple_class_renamed = create_func(SimpleClassRenamed)
    dc = DummyClass()
    with pytest.warns(FutureWarning, match="Argument 'c' is deprecated"):
        dc.add_simple_class_renamed(c=4)
    assert dc.layers[0].a == 4
    with pytest.warns(FutureWarning, match="Argument 'd' is deprecated"):
        dc.add_simple_class_renamed(d=8)
    assert dc.layers[1].b == 8



from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

if TYPE_CHECKING:
    from vispy.color import Colormap


def make_colorbar(
    cmap: 'Colormap', size: tuple[int, int] = (18, 28), horizontal: bool = True
) -> npt.NDArray[np.uint8]:
    """Make a colorbar from a colormap.

    Parameters
    ----------
    cmap : vispy.color.Colormap
        Colormap to create colorbar with.
    size : 2-tuple
        Shape of colorbar.
    horizontal : bool
        If True colobar is oriented horizontal, otherwise it is oriented
        vertical.

    Returns
    -------
    cbar : array
        Array of colorbar in uint8.
    """

    if horizontal:
        basic_values = np.linspace(0, 1, size[1])
        bar = np.tile(np.expand_dims(basic_values, 1), size[0]).transpose(
            (1, 0)
        )
    else:
        basic_values = np.linspace(0, 1, size[0])
        bar = np.tile(np.expand_dims(basic_values, 1), size[1])

    color_array = cmap.map(bar.ravel())
    cbar = color_array.reshape((*bar.shape, 4))

    return np.round(255 * cbar).astype(np.uint8).copy(order='C')



from qtpy.QtWidgets import QDialog

from napari._qt.dialogs.shimmed_plugin_dialog import ShimmedPluginDialog
from napari.settings import get_settings


def test_dialog_accept(qtbot):
    dialog = ShimmedPluginDialog(None, plugins={'plugin1', 'plugin2'})
    qtbot.addWidget(dialog)
    assert dialog.windowTitle() == 'Installed Plugin Warning'
    assert dialog.plugin_text == '\n'.join(sorted({'plugin1', 'plugin2'}))
    assert not dialog.only_new_checkbox.isChecked()
    dialog.okay_btn.click()
    assert dialog.result() == QDialog.DialogCode.Accepted
    assert not get_settings().plugins.only_new_shimmed_plugins_warning


def test_dialog_accept_checked(qtbot):
    dialog = ShimmedPluginDialog(None, plugins={'plugin1', 'plugin2'})
    qtbot.addWidget(dialog)
    assert dialog.plugin_text == '\n'.join(sorted({'plugin1', 'plugin2'}))
    assert not dialog.only_new_checkbox.isChecked()
    dialog.only_new_checkbox.setChecked(True)
    dialog.okay_btn.click()
    assert dialog.result() == QDialog.DialogCode.Accepted
    assert get_settings().plugins.only_new_shimmed_plugins_warning
    assert get_settings().plugins.already_warned_shimmed_plugins == {
        'plugin1',
        'plugin2',
    }


def test_dialog_reject(qtbot):
    dialog = ShimmedPluginDialog(None, plugins={'plugin1', 'plugin2'})
    qtbot.addWidget(dialog)
    assert dialog.plugin_text == '\n'.join(sorted({'plugin1', 'plugin2'}))
    dialog.only_new_checkbox.setChecked(True)
    dialog.reject()
    # rejected doesn't save setting
    assert dialog.result() == QDialog.DialogCode.Rejected
    assert not get_settings().plugins.only_new_shimmed_plugins_warning
    assert get_settings().plugins.already_warned_shimmed_plugins == set()



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np

from napari.layers.base._base_constants import InteractionBoxHandle

if TYPE_CHECKING:
    from napari.layers import Layer


@lru_cache
def generate_interaction_box_vertices(
    top_left: tuple[float, float],
    bot_right: tuple[float, float],
    handles: bool = True,
) -> np.ndarray:
    """
    Generate coordinates for all the handles in InteractionBoxHandle.

    Coordinates are assumed to follow vispy "y down" convention.

    Parameters
    ----------
    top_left : Tuple[float, float]
        Top-left corner of the box
    bot_right : Tuple[float, float]
        Bottom-right corner of the box
    handles : bool
        Whether to also return indices for the transformation handles.

    Returns
    -------
    np.ndarray
        Coordinates of the vertices and handles of the interaction box.
    """
    x0, y0 = top_left
    x1, y1 = bot_right
    vertices = np.array(
        [
            [x0, y0],
            [x0, y1],
            [x1, y0],
            [x1, y1],
        ]
    )

    if handles:
        # add handles at the midpoint of each side
        middle_vertices = np.mean([vertices, vertices[[2, 0, 3, 1]]], axis=0)
        box_height = vertices[0, 1] - vertices[1, 1]
        vertices = np.concatenate([vertices, middle_vertices])

        # add the extra handle for rotation
        extra_vertex = [middle_vertices[0] + [0, box_height * 0.1]]
        vertices = np.concatenate([vertices, extra_vertex])

    return vertices


def generate_transform_box_from_layer(
    layer: Layer, dims_displayed: tuple[int, int]
) -> np.ndarray:
    """
    Generate coordinates for the handles of a layer's transform box.

    Parameters
    ----------
    layer : Layer
        Layer whose transform box to generate.
    dims_displayed : Tuple[int, ...]
        Dimensions currently displayed (must be 2).
    Returns
    -------
    np.ndarray
        Vertices and handles of the interaction box in data coordinates.
    """
    bounds = layer._display_bounding_box_augmented(list(dims_displayed))

    # generates in vispy canvas pos, so invert x and y, and then go back
    top_left, bot_right = (tuple(point) for point in bounds.T[:, ::-1])
    return generate_interaction_box_vertices(
        top_left, bot_right, handles=True
    )[:, ::-1]


def calculate_bounds_from_contained_points(
    points: np.ndarray,
) -> tuple[tuple[float, float], tuple[float, float]]:
    """
    Calculate the top-left and bottom-right corners of an axis-aligned bounding box.

    Parameters
    ----------
    points : np.ndarray
        Array of point coordinates.

    Returns
    -------
    Tuple[Tuple[float, float], Tuple[float, float]]
        Top-left and bottom-right corners of the bounding box.
    """
    if points is None:
        return None

    points = np.atleast_2d(points)
    if points.ndim != 2:
        raise ValueError('only 2D coordinates are accepted')

    x0 = points[:, 0].min()
    x1 = points[:, 0].max()
    y0 = points[:, 1].min()
    y1 = points[:, 1].max()

    return (x0, x1), (y0, y1)


def get_nearby_handle(
    position: np.ndarray, handle_coordinates: np.ndarray
) -> InteractionBoxHandle | None:
    """
    Get the InteractionBoxHandle close to the given position, within tolerance.

    Parameters
    ----------
    position : np.ndarray
        Position to query for.
    handle_coordinates : np.ndarray
        Coordinates of all the handles (except INSIDE).

    Returns
    -------
    Optional[InteractionBoxHandle]
        The nearby handle if any, or InteractionBoxHandle.INSIDE if inside the box.
    """
    top_left = handle_coordinates[InteractionBoxHandle.TOP_LEFT]
    bot_right = handle_coordinates[InteractionBoxHandle.BOTTOM_RIGHT]
    dist = np.linalg.norm(position - handle_coordinates, axis=1)
    tolerance = dist.max() / 100
    close_to_vertex = np.isclose(dist, 0, atol=tolerance)
    if np.any(close_to_vertex):
        idx = int(np.argmax(close_to_vertex))
        return InteractionBoxHandle(idx)
    if np.all((position >= top_left) & (position <= bot_right)):
        return InteractionBoxHandle.INSIDE

    return None
